#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2025/5/19 20:45
# @Author  : 兵
# @email    : 1747193328@qq.com


from NepTrainKit.core import Config
from NepTrainKit.core.structure import table_info
import numpy as np

from vispy import app, scene, visuals
from vispy.geometry import MeshData
from vispy.scene.visuals import Mesh, Line
from vispy.color import Color



class StructurePlotWidget(scene.SceneCanvas):
    def __init__(self, *args, **kwargs):
        super().__init__( *args, **kwargs)
        self.unfreeze()

        self.bgcolor = 'white'  # Set background to white
        self.view = self.central_widget.add_view()
        self.view.camera = 'turntable'  # Interactive camera
        self.ortho = False
        self.atom_items = []  # Store atom meshes and metadata
        self.bond_items = []  # Store bond meshes
        self.lattice_item = None  # Store lattice lines
        self.structure = None
        self.show_bond_flag = False
        self.scale_factor = 1
        initial_camera_dir = (0, -1, 0)  # for a default initialised camera

        self.initial_light_dir = self.view.camera.transform.imap(initial_camera_dir)[:3]

        # Precompute sphere template (reduced resolution)
        phi, theta = np.mgrid[0:np.pi:15j, 0:2 * np.pi:15j]
        x = np.sin(phi) * np.cos(theta)
        y = np.sin(phi) * np.sin(theta)
        z = np.cos(phi)
        self.sphere_vertices = np.c_[x.ravel(), y.ravel(), z.ravel()]
        self.sphere_faces = []
        n_theta = 15
        for i in range(14):
            for j in range(14):
                v0 = i * n_theta + j
                v1 = v0 + 1
                v2 = (i + 1) * n_theta + j
                v3 = v2 + 1
                self.sphere_faces.append([v0, v1, v2])
                self.sphere_faces.append([v1, v3, v2])
        self.sphere_faces = np.array(self.sphere_faces)

        # Precompute cylinder template
        self.n_segments = 8  # Optimized for performance
        theta = np.linspace(0, 2 * np.pi, self.n_segments, endpoint=False)
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        self.cylinder_template = {'cos_theta': cos_theta, 'sin_theta': sin_theta}
        self.cylinder_faces = []
        for i in range(self.n_segments):
            i1, i2 = i, (i + 1) % self.n_segments
            self.cylinder_faces.append([i1, i2, i2 + self.n_segments])
            self.cylinder_faces.append([i1, i2 + self.n_segments, i1 +self.n_segments])
            self.cylinder_faces.append([2 * self.n_segments, i2, i1])  # Bottom cap
            self.cylinder_faces.append([2 * self.n_segments + 1, i1 + self.n_segments, i2 + self.n_segments])  # Top cap
        self.cylinder_faces = np.array(self.cylinder_faces)

    def set_projection(self, ortho=True):
        """Toggle between orthographic and perspective projection."""
        self.ortho = ortho
        current_state = {
            'center': self.view.camera.center,
            'elevation': self.view.camera.elevation,
            'azimuth': self.view.camera.azimuth,

        }
        if self.ortho:
            self.view.camera = scene.cameras.TurntableCamera(
                fov=0,  # Orthographic
                **current_state
            )
        else:
            self.view.camera = scene.cameras.TurntableCamera(
                fov=60,  # Perspective
                **current_state
            )
        self.update()

    def set_show_bonds(self, show_bonds=True):
        """Toggle bond visibility and adjust atom scaling."""
        self.show_bond_flag = show_bonds
        if self.structure is not None:
            self.scale_factor = 0.6 if show_bonds else 1
            self.show_structure(self.structure)

    def update_lighting(self):
        """Update light direction to follow camera."""
        # return
        transform = self.view.camera.transform
        dir = np.concatenate((self.initial_light_dir, [0]))
        light_dir = transform.map(dir)[:3]
        # Update shading filter for atoms, bonds, and halos
        for item in self.atom_items:
            if item['mesh'] and hasattr(item['mesh'], 'shading_filter'):
                item['mesh'].shading_filter.light_dir = tuple(light_dir)
                # item['mesh']._program.draw()  # Force shader update
            if item['halo'] and hasattr(item['halo'], 'shading_filter'):
                item['halo'].shading_filter.light_dir = tuple(light_dir)
                # item['halo']._program.draw()  # Force shader update
        for mesh in self.bond_items:
            if hasattr(mesh, 'shading_filter'):
                mesh.shading_filter.light_dir = tuple(light_dir)
                # mesh._program.draw()  # Force shader update
        self.update()

    def show_lattice(self, structure):
        """Draw the crystal lattice as 12 distinct edges."""
        if self.lattice_item:
            self.lattice_item.parent = None
        origin = np.array([0.0, 0.0, 0.0])
        a1, a2, a3 = structure.cell
        vertices = np.array([
            origin, a1, a2, a1 + a2, a3, a1 + a3, a2 + a3, a1 + a2 + a3
        ])
        edges = [
            [0, 1], [0, 2], [0, 4], [1, 3], [1, 5], [2, 3],
            [2, 6], [3, 7], [4, 5], [4, 6], [5, 7], [6, 7]
        ]
        lines = np.array([vertices[edge] for edge in edges]).reshape(-1, 3)
        self.lattice_item = Line(
            pos=lines,
            color=(0, 0, 0, 1),
            width=1.5,
            connect='segments',
            method='gl',
            parent=self.view.scene
        )

    def show_bond(self, structure):
        """Draw bonds as cylinders between atom pairs."""
        for item in self.bond_items:
            item.parent = None
        self.bond_items = []
        if not self.show_bond_flag:
            return
        bond_pairs = structure.get_bond_pairs()

        # Use precomputed cylinder template
        cos_theta = self.cylinder_template['cos_theta']
        sin_theta = self.cylinder_template['sin_theta']
        cylinder_vertices = []
        cylinder_faces = []
        cylinder_colors = []
        vertex_offset = 0

        for pair in bond_pairs:
            elem0 = str(structure.numbers[pair[0]])
            elem1 = str(structure.numbers[pair[1]])
            pos1 = structure.positions[pair[0]]
            pos2 = structure.positions[pair[1]]
            color1 = Color(table_info.get(elem0, {'color': '#808080'})['color']).rgba
            color2 = Color(table_info.get(elem1, {'color': '#808080'})['color']).rgba
            radius1 = table_info.get(elem0, {'color': '#808080'})['radii'] / 150 * self.scale_factor
            radius2 = table_info.get(elem1, {'radii': 70})['radii'] / 150 * self.scale_factor
            bond_radius = 0.12

            # Bond geometry
            bond_vector = pos2 - pos1
            full_length = np.linalg.norm(bond_vector)
            bond_dir = bond_vector / full_length
            start_point = pos1
            mid_point = pos1 + bond_dir * (radius1 + (full_length - radius1 - radius2) / 2)
            bond1_length = (full_length - radius1 - radius2) / 2 + radius1
            bond2_length = (full_length - radius1 - radius2) / 2 + radius2

            # Orthogonal vectors
            if abs(bond_dir[2]) < 0.999:
                v1 = np.cross(bond_dir, [0, 0, 1])
            else:
                v1 = np.cross(bond_dir, [0, 1, 0])
            v1 = v1 / np.linalg.norm(v1)
            v2 = np.cross(bond_dir, v1)
            v2 = v2 / np.linalg.norm(v2)

            # Cylinder 1
            vertices = []
            colors = []
            for z in [0, bond1_length]:
                for i in range(self.n_segments):
                    x = cos_theta[i] * bond_radius
                    y = sin_theta[i] * bond_radius
                    pos = start_point + bond_dir * z + v1 * x + v2 * y
                    vertices.append(pos)
                    colors.append(color1)
            vertices.append(start_point)  # Bottom center
            colors.append(color1)
            vertices.append(start_point + bond_dir * bond1_length)  # Top center
            colors.append(color1)
            cylinder_vertices.append(vertices)
            cylinder_faces.append(self.cylinder_faces + vertex_offset)
            cylinder_colors.append(colors)
            vertex_offset += len(vertices)

            # Cylinder 2
            vertices = []
            colors = []
            for z in [0, bond2_length]:
                for i in range(self.n_segments):
                    x = cos_theta[i] * bond_radius
                    y = sin_theta[i] * bond_radius
                    pos = mid_point + bond_dir * z + v1 * x + v2 * y
                    vertices.append(pos)
                    colors.append(color2)
            vertices.append(mid_point)  # Bottom center
            colors.append(color2)
            vertices.append(mid_point + bond_dir * bond2_length)  # Top center
            colors.append(color2)
            cylinder_vertices.append(vertices)
            cylinder_faces.append(self.cylinder_faces + vertex_offset)
            cylinder_colors.append(colors)
            vertex_offset += len(vertices)

        # Merge all cylinders
        if cylinder_vertices:
            vertices = np.vstack(cylinder_vertices)
            faces = np.vstack(cylinder_faces)
            colors = np.vstack(cylinder_colors)
            mesh_data = MeshData(vertices=vertices, faces=faces, vertex_colors=colors)
            mesh = Mesh(
                meshdata=mesh_data,
                shading='smooth',
                parent=self.view.scene
            )

            self.bond_items.append(mesh)

    def show_elem(self, structure):
        """Draw atoms as glossy spheres with merged geometry."""
        for item in self.atom_items:
            if item['mesh']:
                item['mesh'].parent = None
            if item['halo']:
                item['halo'].parent = None
        self.atom_items = []

        # Merge all atoms
        all_vertices = []
        all_faces = []
        all_colors = []
        face_offset = 0
        for idx, (n, p) in enumerate(zip(structure.numbers, structure.positions)):
            elem = str(n)
            color = Color(table_info.get(elem, {'color': '#808080'})['color']).rgba
            size = table_info.get(elem, {'radii': 70})['radii'] / 150 * self.scale_factor
            scaled_vertices = self.sphere_vertices * size + p
            all_vertices.append(scaled_vertices)
            all_faces.append(self.sphere_faces + face_offset)
            all_colors.append(np.repeat([color], len(self.sphere_vertices), axis=0))
            face_offset += len(self.sphere_vertices)
            self.atom_items.append({
                'mesh': None,
                'position': p,
                'original_color': color,
                'size': size,
                'halo': None,
                'vertex_range': (len(all_vertices) - 1) * len(self.sphere_vertices)
            })

        # Create single mesh for atoms
        if all_vertices:
            vertices = np.vstack(all_vertices)
            faces = np.vstack(all_faces)
            colors = np.vstack(all_colors)
            mesh_data = MeshData(vertices=vertices, faces=faces, vertex_colors=colors)
            mesh = Mesh(
                meshdata=mesh_data,
                shading='smooth',
                parent=self.view.scene
            )

            for item in self.atom_items:
                item['mesh'] = mesh

        # Highlight bad bonds

        radius_coefficient = Config.getfloat("widget", "radius_coefficient", 0.7)

        bond_pairs = structure.get_bad_bond_pairs(radius_coefficient)
        for pair in bond_pairs:
            self.highlight_atom(pair[0])
            self.highlight_atom(pair[1])

    def highlight_atom(self, atom_index):
        """Highlight an atom with a translucent halo."""
        if 0 <= atom_index < len(self.atom_items):
            atom = self.atom_items[atom_index]
            if atom['halo']:
                atom['halo'].parent = None
            halo_size = atom['size'] * 1.2
            halo_color = [1, 1, 0, 0.6]
            vertices = self.sphere_vertices * halo_size + atom['position']
            mesh_data = MeshData(vertices=vertices, faces=self.sphere_faces)
            halo = Mesh(
                meshdata=mesh_data,
                color=halo_color,
                shading='smooth',
                parent=self.view.scene
            )

            self.atom_items[atom_index]['halo'] = halo
            self.update()

    def reset_atom(self, atom_index):
        """Remove halo from an atom."""
        if 0 <= atom_index < len(self.atom_items):
            atom = self.atom_items[atom_index]
            if atom['halo']:
                atom['halo'].parent = None
                self.atom_items[atom_index]['halo'] = None
            self.update()


    def show_structure(self, structure):
        """Display the entire crystal structure."""
        self.structure = structure


        if self.lattice_item:
            self.lattice_item.parent = None
        self.show_lattice(structure)
        self.show_elem(structure)
        self.show_bond(structure)
        coords = structure.positions
        min_coords = coords.min(axis=0)
        max_coords = coords.max(axis=0)
        center = (min_coords + max_coords) / 2
        size = max_coords - min_coords
        max_dimension = np.max(size)
        fov = 60
        distance = max_dimension / (2 * np.tan(np.radians(fov / 2))) * 2.8
        aspect_ratio = size / np.max(size)
        flat_threshold = 0.5
        if aspect_ratio[0] < flat_threshold and aspect_ratio[1] >= flat_threshold and aspect_ratio[2] >= flat_threshold:
            elevation, azimuth = 0, 0
        elif aspect_ratio[1] < flat_threshold and aspect_ratio[0] >= flat_threshold and aspect_ratio[
            2] >= flat_threshold:
            elevation, azimuth = 0, 0
        elif aspect_ratio[2] < flat_threshold and aspect_ratio[0] >= flat_threshold and aspect_ratio[
            1] >= flat_threshold:
            elevation, azimuth = 90, 0
        else:
            elevation, azimuth = 30, 45
        self.view.camera.set_state({
            'center': tuple(center),
            'elevation': elevation,
            'azimuth': azimuth,

        })
        self.view.camera.distance=distance

        self.update_lighting()


    def on_mouse_move(self, event):
        """Update lighting during rotation."""

        if event.is_dragging:

            self.update_lighting()
if __name__ == '__main__':
    from PySide6.QtWidgets import QApplication
    from NepTrainKit.core.structure import Structure
    app = QApplication([])
    view = StructurePlotWidget()
    view.set_show_bonds(True)
    view.set_projection(True)
    view.show()
    import time
    start = time.time()
    atoms = Structure.read_xyz("good.xyz")
    view.show_structure(atoms)  # 修改为show_structure，与代码一致
    print(time.time() - start)
    QApplication.instance().exec_()


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/20 22:23
# @Author  : 兵
# @email    : 1747193328@qq.com

from .nep import NepResultPlotWidget

from .toolbar import NepDisplayGraphicsToolBar,StructureToolBar



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024/10/17 13:14
# @Author  : 兵
# @email    : 1747193328@qq.com
import platform
import subprocess
import time
import os
import traceback
import re
from collections.abc import Iterable

from PySide6.QtCore import QThread, Signal
from PySide6.QtWidgets import QFileDialog, QApplication
from loguru import logger
from qfluentwidgets import StateToolTip

from NepTrainKit.core import Config
from ase.build.tools import sort as ase_sort
from NepTrainKit.version import UPDATE_EXE, UPDATE_FILE, NepTrainKit_EXE


def timeit(func):
    """
    统计函数耗时
    使用用法：
        @timeit
        def demo():
            pass
    """
    def wrapper(*args, **kwargs):
        start_time = time.time()  # 记录开始时间
        result = func(*args, **kwargs)  # 调用原始函数
        end_time = time.time()  # 记录结束时间
        logger.debug(f"Function '{func.__name__}' executed in {end_time - start_time:.4f} seconds")
        return result
    return wrapper

def check_path_type(path:str) -> str:
    """
    判断路径是文件夹还是文件，即使路径不存在。

    参数:
        path (str): 路径字符串。

    返回:
        str: "folder"（文件夹）、"file"（文件） 。
    """

    if os.path.isdir(path):
        return "folder"
    elif os.path.isfile(path):
        return "file"
    else:
        # 如果路径不存在，进一步检查是否有文件扩展名
        if os.path.splitext(path)[1]:  # 如果有扩展名，可能是文件
            return "file"
        else:  # 否则可能是文件夹
            return "folder"


def call_path_dialog(self,
                     title:str,
                     dialog_type:str = "file",
                     default_filename:str = "",
                     file_filter:str = "",
                     selected_filter:str = "") -> str|None:
    dialog_map = {
        "file": lambda: QFileDialog.getSaveFileName(self, title, os.path.join(Config.get_path(), default_filename), file_filter, selected_filter),
        "select": lambda: QFileDialog.getOpenFileName(self, title, Config.get_path(), file_filter),
        "selects": lambda: QFileDialog.getOpenFileNames(self, title, Config.get_path(), file_filter),
        "directory": lambda: QFileDialog.getExistingDirectory(self, title, Config.get_path())
    }

    dialog_func = dialog_map.get(dialog_type)
    if not dialog_func:
        return None

    select_path = dialog_func()

    if isinstance(select_path, tuple):
        select_path = select_path[0]  # 处理 `getSaveFileName` 和 `getOpenFileName` 返回的 tuple
    elif isinstance(select_path, list):
        if not select_path:
            return None
        select_path = select_path[0]  # `getOpenFileNames` 返回 list

    if not select_path:
        return None

    # 提取目录并保存到配置
    if isinstance(select_path,list):
        path = select_path[0]
    else:
        path = select_path
    if check_path_type(path)=="file":
        last_dir = os.path.dirname(path)
    else:
        last_dir = path

    Config.set("setting", "last_path", last_dir)
    return select_path

def unzip():

    cmd = f"ping -n 3 127.0.0.1&{UPDATE_EXE} {UPDATE_FILE}&ping -n 2 127.0.0.1&start {NepTrainKit_EXE}"

    subprocess.Popen(cmd, shell=True)
    if QApplication.instance():
        QApplication.instance().exit()
    else:
        quit()

class LoadingThread(QThread):
    progressSignal = Signal(int)
    def __init__(self, parent=None, show_tip=True, title='running'):
        super(LoadingThread,self).__init__(parent)
        self.show_tip=show_tip
        self.title=title
        self._parent=parent

    def run(self ):
        result =self._func(*self._args, **self._kwargs)
        if isinstance(result, Iterable):
            for i,_ in enumerate(result):

                self.progressSignal.emit(i)

    def start_work(self,func,*args,**kwargs):
        if self.show_tip:
            self.tip = StateToolTip(self.title, 'Please wait patiently~~', self._parent)
            self.tip.show()
            self.finished.connect(self.__finished_work)
            self.tip.closedSignal.connect(self.quit)
        else:
            self.tip=None
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self.start()

    def __finished_work(self ):
        if self.tip:
            self.tip.setContent('任务完成啦 😆')
            self.tip.setState(True)

    def stop_work(self ):
        self.terminate()




class DataProcessingThread(QThread):
    # 定义信号用于通信
    progressSignal = Signal(int)  # 进度更新信号
    finishSignal = Signal()  # 处理完成信号
    errorSignal = Signal(str)  # 错误信号

    def __init__(self, dataset, process_func):
        super().__init__()
        self.dataset = dataset
        self.process_func = process_func
        self.result_dataset = []

    def run(self):
        """线程主逻辑"""
        try:
            total = len(self.dataset)
            self.progressSignal.emit(0)
            sort_atoms = Config.getboolean("widget", "sort_atoms", False)
            for index, structure in enumerate(self.dataset):
                # 处理每个结构
                processed = self.process_func(structure)
                if sort_atoms:
                    processed = [ase_sort(s) for s in processed]

                self.result_dataset.extend(processed)

                # 发射进度信号 (百分比)
                self.progressSignal.emit(int((index + 1) / total * 100))

            # 处理完成
            self.finishSignal.emit( )
        except Exception as e:
            logger.debug(traceback.format_exc())
            self.errorSignal.emit(str(e))


class FilterProcessingThread(QThread):
    # 定义信号用于通信
    progressSignal = Signal(int)  # 进度更新信号
    finishSignal = Signal()  # 处理完成信号
    errorSignal = Signal(str)  # 错误信号

    def __init__(self,  process_func):
        super().__init__()
        self.process_func = process_func

    def run(self):
        """线程主逻辑"""
        try:
            self.progressSignal.emit(0)
            # 处理每个结构
            self.process_func()
                # 发射进度信号 (百分比)
            self.progressSignal.emit(100)
            # 处理完成
            self.finishSignal.emit( )
        except Exception as e:
            logger.debug(traceback.format_exc())
            self.errorSignal.emit(str(e))


def parse_index_string(s: str, total: int) -> list[int]:
    """Parse an index expression into a list of indices.

    Parameters
    ----------
    s : str
        Index expression like ``"1:10"``, ``":100"`` or ``"::3"``.
        Multiple expressions can be separated by comma or whitespace.
    total : int
        Maximum length of the dataset for bounds checking.

    Returns
    -------
    list[int]
        Sorted list of unique indices within ``range(total)``.
    """
    indices: list[int] = []
    tokens = [t for t in re.split(r"[,\s]+", s.strip()) if t]
    for token in tokens:
        if ":" in token:
            parts = token.split(":")
            if len(parts) > 3:
                continue
            start = int(parts[0]) if parts[0] else None
            end = int(parts[1]) if len(parts) > 1 and parts[1] else None
            step = int(parts[2]) if len(parts) == 3 and parts[2] else None
            slc = slice(start, end, step)
            indices.extend(range(*slc.indices(total)))
        else:
            try:
                idx = int(token)
            except ValueError:
                continue
            if idx < 0:
                idx += total
            if 0 <= idx < total:
                indices.append(idx)
    return sorted(set(indices))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2025/6/18 13:21
# @Author  : 兵
# @email    : 1747193328@qq.com
from PySide6.QtCore import Signal
from PySide6.QtWidgets import QWidget, QVBoxLayout

from NepTrainKit import utils
from NepTrainKit.core import CardManager
from NepTrainKit.custom_widget import MakeDataCardWidget
from NepTrainKit.custom_widget.card_widget import MakeDataCard, FilterDataCard


@CardManager.register_card
class CardGroup(MakeDataCardWidget):
    separator=True
    card_name= "Card Group"
    menu_icon=r":/images/src/images/group.svg"
    #通知下一个card执行
    runFinishedSignal=Signal(int)
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle("Card Group")
        self.setAcceptDrops(True)
        self.index=0
        self.group_widget = QWidget(self)
        # self.setStyleSheet("CardGroup{boder: 2px solid #C0C0C0;}")
        self.viewLayout.addWidget(self.group_widget)
        self.group_layout = QVBoxLayout(self.group_widget)
        self.exportSignal.connect(self.export_data)
        self.windowStateChangedSignal.connect(self.show_card_setting)
        self.filter_widget = QWidget(self)
        self.filter_layout = QVBoxLayout(self.filter_widget)
        self.vBoxLayout.addWidget(self.filter_widget)

        self.filter_card=None
        self.dataset:list=None
        self.result_dataset=[]
        self.resize(400, 200)

    def set_filter_card(self,card):

        self.filter_card=card
        self.filter_layout.addWidget(card)

    def state_changed(self, state):
        super().state_changed(state)
        for card in self.card_list:
            card.state_checkbox.setChecked(state)

    @property
    def card_list(self)->["MakeDataCard"]:

        return [self.group_layout.itemAt(i).widget() for i in range(self.group_layout.count()) ]
    def show_card_setting(self):

        for card in self.card_list:
            card.window_state = self.window_state
            card.windowStateChangedSignal.emit()
    def set_dataset(self,dataset):
        self.dataset =dataset
        self.result_dataset=[]

    def add_card(self, card):
        self.group_layout.addWidget(card)

    def remove_card(self, card):
        self.group_layout.removeWidget(card)

    def clear_cards(self):
        for card in self.card_list:
            self.group_layout.removeWidget(card)

    def closeEvent(self, event):
        for card in self.card_list:
            card.close()
        self.deleteLater()
        super().closeEvent(event)

    def dragEnterEvent(self, event):

        widget = event.source()

        if widget == self:
            return
        if isinstance(widget, (MakeDataCard,CardGroup)):
            event.acceptProposedAction()
        else:
            event.ignore()  # 忽略其他类型的拖拽

    def dropEvent(self, event):

        widget = event.source()
        if widget == self:
            return
        if isinstance(widget, FilterDataCard):
            self.set_filter_card(widget)
        elif isinstance(widget, (MakeDataCard,CardGroup)):
            self.add_card(widget)
        event.acceptProposedAction()

    def on_card_finished(self, index):
        self.run_card_num-=1
        self.card_list[index].runFinishedSignal.disconnect(self.on_card_finished)
        self.result_dataset.extend(self.card_list[index].result_dataset)

        if self.run_card_num==0:
            self.runFinishedSignal.emit(self.index)
            if self.filter_card and self.filter_card.check_state:
                self.filter_card.set_dataset(self.result_dataset)
                self.filter_card.run()

    def stop(self):
        for card in self.card_list:
            card.stop()
        if self.filter_card:
            self.filter_card.stop()

    def run(self):
        # 创建并启动线程
        self.run_card_num = len(self.card_list)

        if self.check_state and self.run_card_num>0:
            self.result_dataset =[]
            for index,card in enumerate(self.card_list):
                if card.check_state:
                    card.set_dataset(self.dataset)
                    card.index=index
                    card.runFinishedSignal.connect(self.on_card_finished)
                    card.run()
                else:
                    self.run_card_num-=1
        else:
            self.result_dataset = self.dataset
            self.runFinishedSignal.emit(self.index)

    def write_result_dataset(self, file,**kwargs):
        if self.filter_card and self.filter_card.check_state:
            self.filter_card.write_result_dataset(file,**kwargs)
            return

        for index,card in enumerate(self.card_list):
            if index==0:
                if "append" not in kwargs:
                    kwargs["append"] = False
            else:
                kwargs["append"] = True
            if card.check_state:
                card.write_result_dataset(file,**kwargs)

    def export_data(self):
        if self.dataset is not None:
            path = utils.call_path_dialog(self, "Choose a file save location", "file",f"export_{self.getTitle()}_structure.xyz")
            if not path:
                return
            thread=utils.LoadingThread(self,show_tip=True,title="Exporting data")
            thread.start_work(self.write_result_dataset, path)
    def to_dict(self):
        data_dict = super().to_dict()

        data_dict["card_list"]=[]

        for card in self.card_list:
            data_dict["card_list"].append(card.to_dict())
        if self.filter_card:
            data_dict["filter_card"]=self.filter_card.to_dict()
        else:
            data_dict["filter_card"]=None

        return data_dict
    def from_dict(self,data_dict):
        self.state_checkbox.setChecked(data_dict['check_state'])
        for sub_card in data_dict.get("card_list",[]):
            card_name=sub_card["class"]
            card  = CardManager.card_info_dict[card_name](self)
            self.add_card(card)
            card.from_dict(sub_card)

        if data_dict.get("filter_card"):
            card_name=data_dict["filter_card"]["class"]
            filter_card  = CardManager.card_info_dict[card_name](self)
            filter_card.from_dict(data_dict["filter_card"])
            self.set_filter_card(filter_card)



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/28 22:45
# @Author  : 兵
# @email    : 1747193328@qq.com
from .layout import FlowLayout
from .label import ProcessLabel
from .completer import CompleterModel, JoinDelegate, ConfigCompleter
from .dialog import (
    GetIntMessageBox,
    SparseMessageBox,
    IndexSelectMessageBox,
    ShiftEnergyMessageBox,
    ProgressDialog,
    PeriodicTableDialog,
)
from .input import SpinBoxUnitInputFrame
from .card_widget import (
    CheckableHeaderCardWidget,
    ShareCheckableHeaderCardWidget,
    MakeDataCardWidget,
)
from .doping_rule import DopingRulesWidget
from .vacancy_rule import VacancyRulesWidget

from .docker import MakeWorkflowArea
from .search_widget import ConfigTypeSearchLineEdit
from .settingscard import MyComboBoxSettingCard, DoubleSpinBoxSettingCard

__all__ = [
    "FlowLayout",
    "ProcessLabel",
    "CompleterModel",
    "JoinDelegate",
    "ConfigCompleter",
    "GetIntMessageBox",
    "SparseMessageBox",
    "IndexSelectMessageBox",
    "ShiftEnergyMessageBox",
    "ProgressDialog",
    "PeriodicTableDialog",
    "SpinBoxUnitInputFrame",
    "CheckableHeaderCardWidget",
    "ShareCheckableHeaderCardWidget",
    "MakeDataCardWidget",
    "MakeWorkflowArea",
    "ConfigTypeSearchLineEdit",
    "MyComboBoxSettingCard",
    "DoubleSpinBoxSettingCard",
    "DopingRulesWidget",
    "VacancyRulesWidget",
]



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/21 19:44
# @Author  : 兵
# @email    : 1747193328@qq.com


from PySide6.QtCore import QUrl, Qt
from PySide6.QtGui import QDesktopServices, QIcon
from PySide6.QtWidgets import QWidget
from qfluentwidgets import FluentIcon as FIF, ScrollArea
from qfluentwidgets import SettingCardGroup, HyperlinkCard, PrimaryPushSettingCard, ExpandLayout, OptionsConfigItem, \
    OptionsValidator, EnumSerializer, SwitchSettingCard

from NepTrainKit.core import Config
from NepTrainKit.custom_widget import MyComboBoxSettingCard, DoubleSpinBoxSettingCard
from NepTrainKit.core.types import ForcesMode, CanvasMode
from NepTrainKit.core.update import UpdateWoker,UpdateNEP89Woker
from NepTrainKit.version import HELP_URL, FEEDBACK_URL, __version__, YEAR, AUTHOR


class SettingsWidget(ScrollArea):
    def __init__(self,parent):

        super().__init__(parent)
        self.setObjectName('SettingsWidget')
        self.scrollWidget = QWidget()

        self.expand_layout = ExpandLayout(self.scrollWidget)

        self.personal_group = SettingCardGroup(
             'Personalization' , self.scrollWidget)


        default_forces = Config.get("widget","forces_data","Raw")
        if default_forces=="Row":
            #没什么用 替换以前的坑 之前写错单词了
            default_forces="Raw"
        self.optimization_forces_card = MyComboBoxSettingCard(
            OptionsConfigItem("forces","forces",ForcesMode(default_forces),OptionsValidator(ForcesMode), EnumSerializer(ForcesMode)),
            FIF.BRUSH,
            'Force data format',
            "Streamline data and speed up drawing",
            texts=[
                "Raw","Norm"
            ],
            default=default_forces,
            parent=self.personal_group
        )
        canvas_type = Config.get("widget","canvas_type","pyqtgraph")

        self.canvas_card = MyComboBoxSettingCard(
            OptionsConfigItem("canvas","canvas",CanvasMode(canvas_type),OptionsValidator(CanvasMode), EnumSerializer(CanvasMode)),
            FIF.BRUSH,
            'Canvas Engine',
            "Choose GPU with vispy",
            texts=[
                "pyqtgraph","vispy"
            ],
            default=canvas_type,
            parent=self.personal_group
        )


        auto_load_config = Config.getboolean("widget","auto_load",False)

        sort_atoms_config = Config.getboolean("widget", "sort_atoms", False)

        self.auto_load_card = SwitchSettingCard(
            QIcon(":/images/src/images/auto_load.svg"),
            self.tr('Auto loading'),
            self.tr('Detect startup path data and load'),

            parent=self.personal_group
        )
        self.auto_load_card.setValue(auto_load_config)

        self.sort_atoms_card = SwitchSettingCard(
            QIcon(":/images/src/images/sort.svg"),
            'Sort atoms',
            'Sort atoms in structures when processing cards',
            parent=self.personal_group
        )
        self.sort_atoms_card.setValue(sort_atoms_config)
        radius_coefficient_config=Config.getfloat("widget","radius_coefficient",0.7)

        self.radius_coefficient_Card = DoubleSpinBoxSettingCard(

            FIF.ALBUM,
            'Covalent radius coefficient',
            'Coefficient used to detect bond length',
            self.personal_group
        )
        self.radius_coefficient_Card.setValue(radius_coefficient_config)
        self.radius_coefficient_Card.setRange(0.0, 1.5)

        self.about_group = SettingCardGroup("About", self.scrollWidget)
        self.help_card = HyperlinkCard(
            HELP_URL,
             'Open Help Page' ,
            FIF.HELP,
             'Help' ,
             'Discover new features and learn useful tips about NepTrainKit.' ,
            self.about_group
        )
        self.feedback_card = PrimaryPushSettingCard(
            "Submit Feedback",
            FIF.FEEDBACK,
            "Submit Feedback",

            'Help us improve NepTrainKit by providing feedback.',
            self.about_group
        )
        self.about_card = PrimaryPushSettingCard(
            'Check for Updates',
            FIF.INFO,
            "About",
            'Copyright ©' + f" {YEAR}, {AUTHOR}. " +
            "Version" + f" {__version__}",
            self.about_group
        )
        self.about_nep89_card = PrimaryPushSettingCard(
            'Check and update',
            FIF.INFO,
            "About NEP89",
            "NEP official NEP89 large model",
            self.about_group
        )
        self.init_layout()
        self.init_signal()
    def init_layout(self):
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        # self.setViewportMargins(0, 80, 0, 20)
        self.setWidget(self.scrollWidget)
        self.setWidgetResizable(True)
        self.scrollWidget.setLayout(self.expand_layout)



        self.personal_group.addSettingCard(self.optimization_forces_card)
        self.personal_group.addSettingCard(self.canvas_card)
        self.personal_group.addSettingCard(self.auto_load_card)
        self.personal_group.addSettingCard(self.radius_coefficient_Card)
        self.personal_group.addSettingCard(self.sort_atoms_card)

        self.about_group.addSettingCard(self.about_nep89_card)
        self.about_group.addSettingCard(self.help_card)
        self.about_group.addSettingCard(self.feedback_card)
        self.about_group.addSettingCard(self.about_card)


        self.expand_layout.addWidget(self.personal_group)
        self.expand_layout.addWidget(self.about_group)

    def init_signal(self):
        self.canvas_card.optionChanged.connect(lambda option:Config.set("widget","canvas_type",option ))
        self.radius_coefficient_Card.valueChanged.connect(lambda value:Config.set("widget","radius_coefficient",value))
        self.optimization_forces_card.optionChanged.connect(lambda option:Config.set("widget","forces_data",option ))
        self.about_card.clicked.connect(self.check_update)
        self.about_nep89_card.clicked.connect(self.check_update_nep89)

        self.auto_load_card.checkedChanged.connect(lambda state:Config.set("widget","auto_load",state))
        self.sort_atoms_card.checkedChanged.connect(lambda state:Config.set("widget","sort_atoms",state))
        # self.about_card.clicked.connect(lambda: QDesktopServices.openUrl(QUrl(RELEASES_URL)))
        self.feedback_card.clicked.connect(
            lambda: QDesktopServices.openUrl(QUrl(FEEDBACK_URL)))

    def check_update(self):
        UpdateWoker(self).check_update()
    def check_update_nep89(self):
        UpdateNEP89Woker(self).check_update()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2025/6/18 13:21
# @Author  : 兵
# @email    : 1747193328@qq.com
import numpy as np
from PySide6.QtWidgets import QFrame, QGridLayout
from qfluentwidgets import BodyLabel, ComboBox, ToolTipFilter, ToolTipPosition, CheckBox

from NepTrainKit.core import CardManager, process_organic_clusters, get_clusters
from NepTrainKit.custom_widget import SpinBoxUnitInputFrame
from NepTrainKit.custom_widget.card_widget import MakeDataCard
from scipy.stats.qmc import Sobol


@CardManager.register_card
class CellScalingCard(MakeDataCard):
    card_name= "Lattice Perturb"
    menu_icon=r":/images/src/images/scaling.svg"
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle("Make Lattice Perturb")

        self.init_ui()

    def init_ui(self):
        self.setObjectName("cell_scaling_card_widget")


        self.engine_label=BodyLabel("Random engine:",self.setting_widget)
        self.engine_type_combo=ComboBox(self.setting_widget)
        self.engine_type_combo.addItem("Sobol")
        self.engine_type_combo.addItem("Uniform")
        self.engine_label.setToolTip("Select random engine")
        self.engine_label.installEventFilter(ToolTipFilter(self.engine_label, 300, ToolTipPosition.TOP))


        self.scaling_condition_frame = SpinBoxUnitInputFrame(self)
        self.scaling_condition_frame.set_input("",1,"float")
        self.scaling_condition_frame.setRange(0,1)
        self.scaling_condition_frame.set_input_value([0.04])

        self.scaling_radio_label=BodyLabel("Max Scaling:",self.setting_widget)
        self.scaling_radio_label.setToolTip("Maximum scaling factor")

        self.scaling_radio_label.installEventFilter(ToolTipFilter(self.scaling_radio_label, 300, ToolTipPosition.TOP))

        self.optional_frame=QFrame(self.setting_widget)
        self.optional_frame_layout = QGridLayout(self.optional_frame)
        self.optional_frame_layout.setContentsMargins(0,0,0,0)
        self.optional_frame_layout.setSpacing(2)
        self.perturb_angle_checkbox=CheckBox( self.setting_widget)
        self.perturb_angle_checkbox.setText("Perturb angle")
        self.perturb_angle_checkbox.setChecked(True)
        self.perturb_angle_checkbox.setToolTip("Also perturb lattice angles")
        self.perturb_angle_checkbox.installEventFilter(ToolTipFilter(self.perturb_angle_checkbox, 300, ToolTipPosition.TOP))


        self.optional_label=BodyLabel("Optional",self.setting_widget)
        self.organic_checkbox=CheckBox("Identify organic", self.setting_widget)
        self.organic_checkbox.setChecked(True)
        self.organic_checkbox.setToolTip("Treat organic molecules as rigid units")
        self.organic_checkbox.installEventFilter(ToolTipFilter(self.organic_checkbox, 300, ToolTipPosition.TOP))

        self.optional_frame_layout.addWidget(self.perturb_angle_checkbox,0,0,1,1)
        self.optional_frame_layout.addWidget(self.organic_checkbox,1,0,1,1)

        self.num_condition_frame = SpinBoxUnitInputFrame(self)
        self.num_condition_frame.set_input("unit",1,"int")
        self.num_condition_frame.setRange(1,10000)
        self.num_label=BodyLabel("Max num:",self.setting_widget)
        self.num_condition_frame.set_input_value([50])
        self.num_label.setToolTip("Number of structures to generate")
        self.num_label.installEventFilter(ToolTipFilter(self.num_label, 300, ToolTipPosition.TOP))

        self.settingLayout.addWidget(self.engine_label,0, 0,1, 1)
        self.settingLayout.addWidget(self.engine_type_combo,0, 1, 1, 2)

        self.settingLayout.addWidget(self.optional_label, 1, 0, 1, 1)
        self.settingLayout.addWidget(self.optional_frame, 1, 1, 1,1)

        self.settingLayout.addWidget(self.scaling_radio_label, 2, 0, 1, 1)

        self.settingLayout.addWidget(self.scaling_condition_frame, 2, 1, 1,2)
        self.settingLayout.addWidget(self.num_label, 3, 0, 1, 1)

        self.settingLayout.addWidget(self.num_condition_frame, 3, 1, 1,2)

    def process_structure(self, structure):
        structure_list=[]
        engine_type=self.engine_type_combo.currentIndex()
        max_scaling=self.scaling_condition_frame.get_input_value()[0]
        max_num=self.num_condition_frame.get_input_value()[0]
        identify_organic=self.organic_checkbox.isChecked()

        if self.perturb_angle_checkbox.isChecked():
            perturb_angles=True
            dim=6 #abc + angles
        else:
            dim=3 #abc
            perturb_angles=False
        if engine_type == 0:

            sobol_engine = Sobol(d=dim, scramble=True)
            sobol_seq = sobol_engine.random(max_num)  # 生成 [0, 1] 的序列
            perturbation_factors = 1 + (sobol_seq - 0.5) * 2 * max_scaling
        else:
            perturbation_factors = 1 + np.random.uniform(-max_scaling, max_scaling, (max_num, dim))

        orig_lattice = structure.cell.array
        orig_lengths = np.linalg.norm(orig_lattice, axis=1)
        unit_vectors = orig_lattice / orig_lengths[:, np.newaxis]  # 原始晶格的单位向量
        if identify_organic:
            clusters, is_organic_list = get_clusters(structure)
        for i in range(max_num):
            # 提取微扰因子
            new_structure=structure.copy()
            length_factors = perturbation_factors[i, :3]
            new_lengths = orig_lengths * length_factors

            # 构造新晶格：仅缩放长度，保持方向
            new_lattice = unit_vectors * new_lengths[:, np.newaxis]

            # 可选：扰动角度
            if perturb_angles:
                angle_factors = perturbation_factors[i, 3:]
                angles = np.arccos([
                    np.dot(orig_lattice[1], orig_lattice[2]) / (orig_lengths[1] * orig_lengths[2]),
                    np.dot(orig_lattice[0], orig_lattice[2]) / (orig_lengths[0] * orig_lengths[2]),
                    np.dot(orig_lattice[0], orig_lattice[1]) / (orig_lengths[0] * orig_lengths[1])
                ])
                new_angles = angles * angle_factors
                # 重新构造晶格（保持角度扰动）
                new_lattice = np.zeros((3, 3), dtype=np.float32)
                new_lattice[0] = [new_lengths[0], 0, 0]
                new_lattice[1] = [new_lengths[1] * np.cos(new_angles[2]),
                                  new_lengths[1] * np.sin(new_angles[2]), 0]
                cx = new_lengths[2] * np.cos(new_angles[1])
                cy = new_lengths[2] * (np.cos(new_angles[0]) - np.cos(new_angles[1]) * np.cos(new_angles[2])) / np.sin(
                    new_angles[2])
                cz = np.sqrt(max(new_lengths[2] ** 2 - cx ** 2 - cy ** 2, 0))  # 防止负值
                new_lattice[2] = [cx, cy, cz]

            # 缩放原子位置

            new_structure.info["Config_type"] = new_structure.info.get("Config_type","") + f" Scaling(scaling={max_scaling},{'uniform' if engine_type==1 else 'Sobol'  })"

            new_structure.set_cell(new_lattice,  scale_atoms=True)
            if identify_organic:
                process_organic_clusters(structure, new_structure, clusters, is_organic_list )

            structure_list.append(new_structure)
        return structure_list

    def to_dict(self):
        data_dict = super().to_dict()

        data_dict['engine_type'] = self.engine_type_combo.currentIndex()
        data_dict['perturb_angle'] = self.perturb_angle_checkbox.isChecked()
        data_dict['organic'] = self.organic_checkbox.isChecked()

        data_dict['scaling_condition'] = self.scaling_condition_frame.get_input_value()
        data_dict['num_condition'] = self.num_condition_frame.get_input_value()
        return data_dict

    def from_dict(self, data_dict):
        super().from_dict(data_dict)

        self.organic_checkbox.setChecked(data_dict.get("organic", False))
        self.engine_type_combo.setCurrentIndex(data_dict['engine_type'])
        self.perturb_angle_checkbox.setChecked(data_dict['perturb_angle'])
        self.scaling_condition_frame.set_input_value(data_dict['scaling_condition'])
        self.num_condition_frame.set_input_value(data_dict['num_condition'])



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/12/20 17:18
# @Author  : 兵
# @email    : 1747193328@qq.com
import json
import os.path

import numpy as np
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QWidget, QGridLayout, QApplication

from NepTrainKit.core import MessageManager, CardManager
from NepTrainKit.custom_widget import MakeWorkflowArea

from NepTrainKit.views.cards import   ConsoleWidget


from NepTrainKit.version import __version__
from NepTrainKit import utils
from ase.io import read as ase_read



class MakeDataWidget(QWidget):
    """
微扰训练集制作
    """

    def __init__(self,parent=None):
        super().__init__(parent)
        self._parent = parent
        self.setObjectName("MakeDataWidget")
        self.setAcceptDrops(True)
        self.nep_result_data=None
        self.init_action()
        self.init_ui()
        self.dataset=None


    def dragEnterEvent(self, event):

        # 检查拖拽的内容是否包含文件
        if event.mimeData().hasUrls():
            event.acceptProposedAction()  # 接受拖拽事件

        else:
            event.ignore()  # 忽略其他类型的拖拽

    def dropEvent(self, event):
        # 获取拖拽的文件路径
        # print("dropEvent",event)
        urls = event.mimeData().urls()

        if urls:
            # 获取第一个文件路径
            structures_path = []
            for url in urls:
                file_path = url.toLocalFile()
                if (file_path.endswith(".xyz") or
                    file_path.endswith(".vasp") or
                    file_path.endswith(".cif")):
                    structures_path.append(file_path)

                elif file_path.endswith(".json"):
                    self.parse_card_config(file_path)
                else:
                    MessageManager.send_info_message("Only .xyz .vasp .cif or json files are supported for import.")
            if structures_path:
                self.load_base_structure(structures_path)

        # event.accept()

    def showEvent(self, event):
        if hasattr(self._parent,"load_menu"):
            self._parent.load_menu.addAction(self.load_card_config_action)
        if hasattr(self._parent,"save_menu"):
            self._parent.save_menu.addAction(self.export_card_config_action)

    def hideEvent(self, event):
        if hasattr(self._parent,"load_menu"):
            self._parent.load_menu.removeAction(self.load_card_config_action)
        if hasattr(self._parent,"save_menu"):
            self._parent.save_menu.removeAction(self.export_card_config_action)

    def init_action(self):
        self.export_card_config_action = QAction(QIcon(r":/images/src/images/save.svg"), "Export Card Config")
        self.export_card_config_action.triggered.connect(self.export_card_config)
        self.load_card_config_action = QAction(QIcon(r":/images/src/images/open.svg"), "Import Card Config")
        self.load_card_config_action.triggered.connect(self.load_card_config)
    def init_ui(self):

        self.gridLayout = QGridLayout(self)
        self.gridLayout.setObjectName("make_data_gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.workspace_card_widget = MakeWorkflowArea(self)
        self.setting_group=ConsoleWidget(self)
        self.setting_group.runSignal.connect(self.run_card)
        self.setting_group.stopSignal.connect(self.stop_run_card)
        self.setting_group.newCardSignal.connect(self.add_card)
        self.gridLayout.addWidget(self.setting_group, 0, 0, 1, 1)
        self.gridLayout.addWidget(self.workspace_card_widget, 1, 0, 1, 1)
        self.setLayout(self.gridLayout)

    def load_base_structure(self,paths):

        structures_list = []
        for path  in paths:
            atoms = ase_read(path,":")
            #ase有时候会将字符串解析成数组或者int  这里转换成str
            for atom in atoms:

                if 'config_type' in atom.info:
                    atom.info["Config_type"]=atom.info["config_type"]
                    del atom.info["config_type"]



                if isinstance(atom.info.get("Config_type"),np.ndarray):
                    if atom.info["Config_type"].size==0:

                        atom.info["Config_type"] = ""
                    else:
                        atom.info["Config_type"]=" ".join(atom.info["Config_type"])

                else:
                    atom.info["Config_type"]=str(atom.info.get("Config_type",""))
                structures_list.append(atom)

        self.dataset=structures_list
        MessageManager.send_success_message(f"success load {len(structures_list)} structures.")

    def open_file(self):
        path = utils.call_path_dialog(self,"Please choose the structure files",
                                      "selects",file_filter="Structure Files (*.xyz *.vasp *.cif)")

        if path:
            self.load_base_structure(path)
    def _export_file(self,path):
        if os.path.exists(path):
            os.remove(path)
        with open(path, "w") as file:
            for card in self.workspace_card_widget.cards:
                if card.check_state:

                    card.write_result_dataset(file,append=True)


    def export_file(self):


        path = utils.call_path_dialog(self, "Choose a file save location", "file",default_filename="make_dataset.xyz")
        if path:
            thread = utils.LoadingThread(self, show_tip=True, title="Exporting data")
            thread.start_work(self._export_file, path)
    def run_card(self):
        if not  self.dataset  :
            MessageManager.send_info_message("Please import the structure file first. You can drag it in directly or import it from the upper left corner!")
            return
        self.stop_run_card()
        first_card=self._next_card(-1)
        if first_card:
            first_card.dataset = self.dataset

            first_card.runFinishedSignal.connect(self._run_next_card)
            first_card.run()
        else:
            MessageManager.send_info_message("No card selected. Please select a card in the workspace.")
    def _next_card(self,current_card_index=-1):

        cards=self.workspace_card_widget.cards
        if current_card_index+1 >=len(cards):
            return None
        current_card_index+=1
        for i,card in enumerate(cards[current_card_index:]):

            if card.check_state:
                card.index=i+current_card_index
                return card
            else:
                continue
        return None
    def _run_next_card(self,current_card_index):

        cards=self.workspace_card_widget.cards
        current_card=cards[current_card_index]
        current_card.runFinishedSignal.disconnect(self._run_next_card)

        next_card=self._next_card(current_card_index )
        if current_card.result_dataset and next_card:
            next_card.set_dataset(current_card.result_dataset)

            next_card.runFinishedSignal.connect(self._run_next_card)
            next_card.run()
        else:
            MessageManager.send_success_message("Perturbation training set created successfully.")
    def stop_run_card(self):
        for card in self.workspace_card_widget.cards:
            card.runFinishedSignal.disconnect(self._run_next_card)
            card.stop()

    def add_card(self,card_name):


        if card_name not in CardManager.card_info_dict:
            MessageManager.send_warning_message("no card")
            return
        card=CardManager.card_info_dict[card_name](self)
        self.workspace_card_widget.add_card(card)
        return card

    def export_card_config(self):
        cards=self.workspace_card_widget.cards
        if not cards:
            MessageManager.send_warning_message("No cards in workspace.")

            return

        path = utils.call_path_dialog(self, "Choose a file save location", "file", default_filename="card_config.json")
        if path:
            config={}
            config["software_version"]=__version__
            config["cards"]=[]
            for card in cards:
                config["cards"].append(card.to_dict())


            with open(path, "w",encoding="utf-8") as file:
                json.dump(config, file, indent=4,ensure_ascii=False)
            MessageManager.send_success_message("Card configuration exported successfully.")
    def load_card_config(self):
        path = utils.call_path_dialog(self, "Choose a card configuration file", "select" )
        if path:

            self.parse_card_config(path)
    def parse_card_config(self,path):
            try:
                with open(path, "r",encoding="utf-8") as file:
                    config = json.load(file)
            except:
                MessageManager.send_warning_message("Invalid card configuration file.")
                return
            self.workspace_card_widget.clear_cards()
            cards=config.get("cards")
            for card in cards:
                name=card.get("class")
                card_widget=self.add_card(name)
                card_widget.from_dict(card)


if __name__ == "__main__":
    import sys
    app = QApplication(sys.argv)

    window = MakeDataWidget()
    window.resize( 800,600)
    window.show()
    sys.exit(app.exec_())


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/21 14:45
# @Author  : 兵
# @email    : 1747193328@qq.com
import json
import os
import re
from copy import deepcopy
import numpy as np
from ase import neighborlist
from ase.geometry import find_mic
from scipy.sparse.csgraph import connected_components

from NepTrainKit import utils, module_path

atomic_numbers={ 'H': 1, 'He': 2, 'Li': 3, 'Be': 4,
                 'B': 5, 'C': 6, 'N': 7, 'O': 8,
                 'F': 9, 'Ne': 10, 'Na': 11, 'Mg': 12,
                 'Al': 13, 'Si': 14, 'P': 15, 'S': 16,
                 'Cl': 17, 'Ar': 18, 'K': 19, 'Ca': 20,
                 'Sc': 21, 'Ti': 22, 'V': 23, 'Cr': 24,
                 'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28,
                 'Cu': 29, 'Zn': 30, 'Ga': 31, 'Ge': 32,
                 'As': 33, 'Se': 34, 'Br': 35, 'Kr': 36,
                 'Rb': 37, 'Sr': 38, 'Y': 39, 'Zr': 40,
                 'Nb': 41, 'Mo': 42, 'Tc': 43, 'Ru': 44,
                 'Rh': 45, 'Pd': 46, 'Ag': 47, 'Cd': 48,
                 'In': 49, 'Sn': 50, 'Sb': 51, 'Te': 52, 'I': 53,
                 'Xe': 54, 'Cs': 55, 'Ba': 56, 'La': 57, 'Ce': 58,
                 'Pr': 59, 'Nd': 60, 'Pm': 61, 'Sm': 62, 'Eu': 63,
                 'Gd': 64, 'Tb': 65, 'Dy': 66, 'Ho': 67, 'Er': 68,
                 'Tm': 69, 'Yb': 70, 'Lu': 71, 'Hf': 72, 'Ta': 73,
                 'W': 74, 'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78,
                 'Au': 79, 'Hg': 80, 'Tl': 81, 'Pb': 82, 'Bi': 83,
                 'Po': 84, 'At': 85, 'Rn': 86, 'Fr': 87, 'Ra': 88,
                 'Ac': 89, 'Th': 90, 'Pa': 91, 'U': 92, 'Np': 93,
                 'Pu': 94, 'Am': 95, 'Cm': 96, 'Bk': 97, 'Cf': 98,
                 'Es': 99, 'Fm': 100, 'Md': 101, 'No': 102, 'Lr': 103,
                 'Rf': 104, 'Db': 105, 'Sg': 106, 'Bh': 107, 'Hs': 108,
                 'Mt': 109, 'Ds': 110, 'Rg': 111, 'Cn': 112, 'Nh': 113,
                 'Fl': 114, 'Mc': 115, 'Lv': 116, 'Ts': 117, 'Og': 118}

with open(os.path.join(module_path, "Config/ptable.json"), "r", encoding="utf-8") as f:
    table_info = json.loads(f.read())

class Structure:
    """
    extxyz格式的结构类
    原子坐标是笛卡尔坐标
    """
    def __init__(self, lattice, structure_info, properties, additional_fields):
        super().__init__()
        self.properties = properties
        self.lattice = np.array(lattice,dtype=np.float32).reshape((3,3))  # Optional: Lattice vectors
        self.structure_info = structure_info
        self.additional_fields = additional_fields
        if "Config_type" not in self.additional_fields.keys():
            self.additional_fields["Config_type"] = ""
        if "forces" in self.structure_info.keys():
            self.force_label="forces"
        else:
            self.force_label = "force"

    def __len__(self):
        return len(self.elements)

    @classmethod
    def read_xyz(cls, filename):
        with open(filename, 'r') as f:
            structure = cls.parse_xyz(f.read())
        return structure
    @property
    def cell(self):
        return self.lattice

    @property
    def volume(self):
        return np.abs(np.linalg.det(self.lattice))

    @property
    def numbers(self):
        return [atomic_numbers[element] for element in self.elements]

    @property
    def formula(self):
        #这种形式会导致化学式过长 比如有机分子环境下
        # diffs = np.diff(self.numbers)
        # # 找到变化的地方，包含第一个元素的起始位置
        # change_points = np.where(diffs != 0)[0] + 1
        # # 在变化的位置之间进行分段
        # segments = np.split(self.elements, change_points)
        # # 格式化每个段落
        # result = [f"{segment[0]}{len(segment)}" for segment in segments]
        # return "".join(result)
        # 改成下面的形式
        formula = ""
        elems={}
        for element in self.elements:
            if element in elems.keys():
                elems[element]+=1
            else:
                elems[element]=1
        for element,count in elems.items():
            formula+=element+str(count)
        return formula

    @property
    def html_formula(self):
        formula = ""
        elems = {}
        for element in self.elements:
            if element in elems.keys():
                elems[element] += 1
            else:
                elems[element] = 1
        for element, count in elems.items():
            formula += element +"<sub>" + str(count) + "</sub>"
        return formula

    @property
    def per_atom_energy(self):
        return self.energy/self.num_atoms
    @property
    def energy(self):
        return self.additional_fields["energy"]
    @energy.setter
    def energy(self,new_energy):
        self.additional_fields["energy"] = new_energy
    @property
    def forces(self):
        return self.structure_info[self.force_label]

    @property
    def nep_virial(self):
        try:
            vir=np.array(self.virial.split(" "),dtype=np.float32)
        except:
            #检查下有没有压强
            try:
                vir= np.array(self.stress.split(" "),dtype=np.float32)*self.volume*-1
            except:
                raise ValueError("No virial or stress data")

        return vir[[0,4,8,1,5,6]]/self.num_atoms

    @property
    def nep_dipole(self):
        dipole=np.array(self.dipole.split(" "),dtype=np.float32)
        return dipole/self.num_atoms

    @property
    def nep_polarizability(self):
        vir = np.array(self.pol.split(" "), dtype=np.float32)
        return vir[[0,4,8,1,5,6]] / self.num_atoms

    def get_chemical_symbols(self):
        return self.elements

    @property
    def elements(self):
        return self.structure_info['species']

    @property
    def positions(self):
        return self.structure_info['pos']

    @property
    def num_atoms(self):
        return len(self.elements)

    def copy(self):
        return deepcopy(self)

    def set_lattice(self, new_lattice: np.ndarray,in_place=False):
        """
        根据新晶格缩放原子位置，支持原地修改或返回新对象。

        :param new_lattice: 新晶格矩阵（3x3 numpy 数组）
        :param in_place: 是否修改当前对象（默认 False，返回新对象）
        :return: 更新后的 Structure 对象（若 in_place=True，则返回 self）
        """
        target = self if in_place else self.copy()
        old_lattice = target.lattice
        old_positions = target.positions

        # 计算变换矩阵（参考 ASE）
        M = np.linalg.solve(old_lattice, new_lattice)
        new_positions = old_positions @ M

        # 更新晶格和坐标
        target.lattice = new_lattice
        target.structure_info['pos'] = new_positions

        return target

    def supercell(self, scale_factor, order="atom-major", tol=1e-5):
        """
        按指定比例因子扩展晶胞，参考 ASE 的高效实现。
        保持晶格角度不变，并支持按元素排序。

        :param scale_factor: 扩展比例因子（标量或长度为3的数组，对应 a、b、c 方向）
        :param order: 原子排序方式，"cell-major"（默认）或 "atom-major"
        :param tol: 数值容差，用于边界检查
        :return: 扩展后的新 Structure 对象
        :raises ValueError: 如果 scale_factor 无效
        """
        # 输入验证和标准化
        scale_factor = np.asarray(scale_factor, dtype=np.float32)
        if scale_factor.size == 1:
            scale_factor = np.full(3, scale_factor)
        if scale_factor.size != 3:
            raise ValueError("scale_factor 必须是标量或长度为3的数组")
        if scale_factor.min() < 1:
            raise ValueError("scale_factor 必须大于等于 1")

        # 输入验证（同上）
        scale_factor = np.asarray(scale_factor, dtype=np.int64)  # 限制为整数扩展

        # 计算新晶格（各方向独立扩展）
        new_lattice = self.lattice * scale_factor[:, None]

        # 转换原始坐标到分数坐标并包裹
        inv_orig_lattice = np.linalg.inv(self.lattice)
        frac_pos = self.positions @ inv_orig_lattice
        frac_pos = frac_pos % 1.0  # 严格包裹

        # 生成扩展网格（明确轴向顺序）
        n_a, n_b, n_c = scale_factor
        # 生成平移向量时确保a方向优先
        offsets_a = np.arange(n_a)[:, None] * np.array([1, 0, 0])  # a方向偏移
        offsets_b = np.arange(n_b)[:, None] * np.array([0, 1, 0])  # b方向偏移
        offsets_c = np.arange(n_c)[:, None] * np.array([0, 0, 1])  # c方向偏移

        # 计算所有偏移组合（a方向最外层循环）
        full_offsets = (offsets_a[:, None, None] +
                        offsets_b[None, :, None] +
                        offsets_c[None, None, :]).reshape(-1, 3)

        # 扩展分数坐标
        expanded_frac = frac_pos[:, None, :] + full_offsets[None, :, :]
        expanded_frac = expanded_frac.reshape(-1, 3) / scale_factor  # 归一化到新分数坐标

        # 转换到新笛卡尔坐标
        new_positions = expanded_frac @ new_lattice

        # 元素扩展（保持a方向优先顺序）
        if order == "cell-major":
            new_elements = np.tile(self.elements, np.prod(scale_factor))
        elif order == "atom-major":
            new_elements = np.repeat(self.elements, np.prod(scale_factor))
        else:
            raise ValueError( )

        # 更新结构信息
        structure_info = {}
        structure_info['pos'] = new_positions.astype(np.float32)
        structure_info['species'] = new_elements

        properties=[{'name': 'species', 'type': 'S', 'count': 1}, {'name': 'pos', 'type': 'R', 'count': 3}]
        # 设置周期性边界条件（假设与原始一致）
        additional_fields={}
        additional_fields['pbc'] = self.additional_fields.get('pbc', "T T T")
        additional_fields["Config_type"] =self.additional_fields.get('Config_type', "")+f" super cell({scale_factor})"

        return Structure(new_lattice, structure_info, properties, additional_fields)
    def adjust_reasonable(self, coefficient=0.7):
        """
        根据传入系数 对比共价半径和实际键长，
        如果实际键长小于coefficient*共价半径之和，判定为不合理结构 返回False
        否则返回 True
        :param coefficient: 系数
        :return:

        """
        distance_info = self.get_mini_distance_info()
        for elems, bond_length in distance_info.items():
            elem0_info = table_info[str(atomic_numbers[elems[0]])]
            elem1_info = table_info[str(atomic_numbers[elems[1]])]

            # 相邻原子距离小于共价半径之和×系数就选中
            if (elem0_info["radii"] + elem1_info["radii"]) * coefficient > bond_length * 100:
                return False
        return True





    # 在序列化时使用 __getstate__ 进行处理
    def __getstate__(self):
        # 返回对象的状态字典，这里可以控制哪些属性需要序列化

        state = self.__dict__.copy()

        return state

    # 反序列化时使用 __setstate__
    def __setstate__(self, state):
        self.__dict__.update(state)

    def __getattr__(self, item):

        if item in self.additional_fields.keys():
            return self.additional_fields[item]
        elif item in self.structure_info.keys():
            return self.structure_info[item]
        else:
            raise AttributeError


    @classmethod
    def parse_xyz(cls, lines):
        """
        Parse a single structure from a list of lines.
        """
        if isinstance(lines, str):
            lines = lines.strip().split('\n')
        # Parse the second line (global properties)
        global_properties = lines[1].strip()
        lattice, properties, additional_fields = cls._parse_global_properties(global_properties)
        array = np.array([line.split() for line in lines[2:]],dtype=object )

        structure_info = {}
        index = 0

        for prop in properties:

            _info = array[:, index:index + prop["count"]]
            #
            # _info =[row[index:index + prop["count"]] for row in array]

            if prop["type"] == "S":
                pass

            elif prop["type"] == "R":
                _info=_info.astype( np.float32)

            else:
                pass
            if prop["count"] == 1:
                _info = _info.ravel()
            else:
                _info = _info.reshape((-1, prop["count"]))

            structure_info[prop["name"]] = _info
            index += prop["count"]

        # return
        return cls(lattice, structure_info, properties, additional_fields)

    @classmethod
    def _parse_global_properties(cls, line):
        """
        Parse global properties from the second line of an XYZ block.
        """
        pattern = r'(\w+)=\s*"([^"]+)"|(\w+)=\s*([\S]+)'
        matches = re.findall(pattern, line)
        properties = []
        lattice = None
        additional_fields = {}

        for match in matches:
            key = match[0] or match[2]
            # key=key.capitalize()
            value = match[1] or match[3]

            if key.capitalize()  == "Lattice":
                lattice = list(map(float, value.split()))
            elif key.capitalize()  == "Properties":
                # Parse Properties details
                properties = cls._parse_properties(value)
            else:

                if '"' in value:

                    value = value.strip('"')  # 去掉引号
                else:
                    try:
                        value = float(value)
                    except Exception as e:
                        value = value
                if key == "config_type" or key == "Config_type":
                    # 这里是为了后面的Config搜索做统一
                    key = "Config_type"
                    value=str(value)
                if key.lower() in ("energy", "pbc","virial","stress"):
                    key=key.lower()
                additional_fields[key] = value
                # print(additional_fields)
        return lattice, properties, additional_fields

    @staticmethod
    def _parse_properties(properties_str):
        """
        Parse `Properties` attribute string to extract atom-specific fields.
        """
        tokens = properties_str.split(":")
        parsed_properties = []
        i = 0
        while i < len(tokens):
            name = tokens[i]
            dtype = tokens[i + 1]
            count = int(tokens[i + 2]) if i + 2 < len(tokens) else 1
            parsed_properties.append({"name": name, "type": dtype, "count": count})
            i += 3
        return parsed_properties

    @staticmethod
    @utils.timeit
    def read_multiple(filename ):
        """
        Read a multi-structure XYZ file and return a list of Structure objects.
        """


        # data_to_process = []
        structures = []

        with open(filename, 'r') as file:
            lines = file.read().splitlines()
        i = 0
        while i < len(lines):
            num_atoms = lines[i].strip()

            if not num_atoms:
                i += 1
                continue
            num_atoms = int(num_atoms)
            end = i + 2 + num_atoms
            structure_lines = lines[i:end]

            structure = Structure.parse_xyz(structure_lines)
            structures.append(structure)
            i = end
        # with multiprocessing.Pool(processes=multiprocessing.cpu_count()) as pool:
            # map 函数并行处理每个结构的读取
            # structures = pool.map(Structure.read, data_to_process)

        return structures

    def write(self, file):
        """
        Write the current structure to an XYZ file.
        """

        # Write number of atoms
        file.write(f"{self.num_atoms}\n")

        # Write global properties
        global_line = []
        if self.lattice.size!=0:
            global_line.append(f'Lattice="' + ' '.join(f"{x}" for x in self.cell.flatten()) + '"')

        props = ":".join(f"{p['name']}:{p['type']}:{p['count']}" for p in self.properties)
        global_line.append(f"Properties={props}")
        for key, value in self.additional_fields.items():

            if isinstance(value, (float, int)):
                global_line.append(f"{key}={value}")

            else:
                global_line.append(f'{key}="{value}"')
        file.write(" ".join(global_line) + "\n")

        for row in range(self.num_atoms):
            line = ""
            for prop  in self.properties :
                if prop["count"] == 1:
                    values=[self.structure_info[prop["name"]][row]]
                else:
                    values=self.structure_info[prop["name"]][row,:]



                if prop["type"] == 'S':  # 字符串类型
                    line += " ".join([f"{x }" for x in values]) + " "

                elif prop["type"] == 'R':  # 浮点数类型
                    line += " ".join([f"{x:.10g}" for x in values]) + " "
                else:
                    line += " ".join([f"{x}" for x in values]) + " "
            file.write(line.strip() + "\n")

    def get_all_distances(self):
        return  calculate_pairwise_distances(self.cell, self.positions,False)

    def get_mini_distance_info(self):
        """
        返回原子对之间的最小距离
        """
        dist_matrix = calculate_pairwise_distances(self.cell, self.positions,False)
        symbols=self.elements
        # 提取上三角矩阵（排除对角线）
        i, j = np.triu_indices(len(self), k=1)
        # 用字典来存储每种元素对的最小键长
        bond_lengths = {}
        # 遍历所有原子对，计算每一对元素的最小键长
        for idx in range(len(i)):
            atom_i, atom_j = symbols[i[idx]], symbols[j[idx]]
            # if atom_i==atom_j:
            #     continue
            # 获取当前键长
            bond_length = dist_matrix[i[idx], j[idx]]
            # if bond_length>5:
            #     continue
            # 确保元素对按字母顺序排列，避免 Cs-Ag 和 Ag-Cs 视为不同
            element_pair = tuple(sorted([atom_i, atom_j]))
            # 如果该元素对尚未存在于字典中，初始化其最小键长
            if element_pair not in bond_lengths:
                bond_lengths[element_pair] = bond_length
            else:
                # 更新最小键长
                bond_lengths[element_pair] = min(bond_lengths[element_pair], bond_length)

        return bond_lengths
    def get_bond_pairs(self):
        """
        返回在范围内的所有键长
        """
        i, j = np.triu_indices(len(self), k=1)
        pos = np.array(self.positions)
        diff = pos[i] - pos[j]
        upper_distances = np.linalg.norm(diff, axis=1)
        covalent_radii = np.array([table_info[str(n)]["radii"] / 100 for n in self.numbers])
        radius_sum = covalent_radii[i] + covalent_radii[j]
        bond_mask = (upper_distances < radius_sum * 1.15)
        bond_pairs = [(i[k], j[k]) for k in np.where(bond_mask)[0]]
        return bond_pairs

    def get_bad_bond_pairs(self, coefficient=0.8):
        """
        根据键长阈值判断
        返回所有的非物理键长
        """
        i, j = np.triu_indices(len(self), k=1)
        distances = self.get_all_distances()
        upper_distances = distances[i, j]
        covalent_radii = np.array([table_info[str(n)]["radii"] / 100 for n in self.numbers])
        radius_sum = covalent_radii[i] + covalent_radii[j]
        bond_mask = (upper_distances < radius_sum * coefficient)

        bad_bond_pairs = [(i[k], j[k]) for k in np.where(bond_mask)[0]]
        return bad_bond_pairs

def calculate_pairwise_distances(lattice_params:np.ndarray, atom_coords:np.ndarray, fractional=True):
    """
    计算晶体中所有原子对之间的距离，考虑周期性边界条件

    参数:
    lattice_params: 晶格参数，3x3 numpy array 表示晶格向量 (a, b, c)
    atom_coords: 原子坐标，Nx3 numpy array
    fractional: 是否为分数坐标 (True) 或笛卡尔坐标 (False)

    返回:
    distances: NxN numpy array，所有原子对之间的距离
    """


    if fractional:
        atom_coords = np.dot(atom_coords, lattice_params)

    diff = atom_coords[np.newaxis, :, :] - atom_coords[:, np.newaxis, :]
    shifts = np.array(np.meshgrid([-1, 0, 1], [-1, 0, 1], [-1, 0, 1]), dtype=np.int8).T.reshape(-1, 3)
    lattice_shifts = np.dot(shifts, lattice_params)
    all_diffs = diff[:, :, np.newaxis, :] + lattice_shifts[np.newaxis, np.newaxis, :, :]
    all_distances = np.sqrt(np.sum(all_diffs ** 2, axis=-1))
    distances = np.min(all_distances, axis=-1)
    np.fill_diagonal(distances, 0)
    return distances



# 判断团簇是否为有机分子
def is_organic_cluster(symbols):
    has_carbon = 'C' in symbols
    organic_elements = {'H', 'O', 'N', 'S', 'P'}
    has_organic_elements = any(symbol in organic_elements for symbol in symbols)
    return has_carbon and has_organic_elements

# 识别结构中的团簇
def get_clusters(structure):
    cutoff = neighborlist.natural_cutoffs(structure)
    nl = neighborlist.NeighborList(cutoff, self_interaction=False, bothways=True)
    nl.update(structure)
    matrix = nl.get_connectivity_matrix()
    n_components, component_list = connected_components(matrix)

    clusters = []
    is_organic_list = []
    for i in range(n_components):
        cluster_indices = [j for j in range(len(structure)) if component_list[j] == i]
        cluster_symbols = [structure[j].symbol for j in cluster_indices]
        clusters.append(cluster_indices)
        is_organic_list.append(is_organic_cluster(cluster_symbols))
    return clusters, is_organic_list

# 解包跨越边界的分子
def unwrap_molecule(structure, cluster_indices):
    pos = structure.positions[cluster_indices]
    cell = structure.cell
    ref_pos = pos[0]
    unwrapped_pos = [ref_pos]

    for i in range(1, len(cluster_indices)):
        delta = pos[i] - ref_pos
        mic_delta, _ = find_mic(delta, cell, pbc=True)
        unwrapped_pos.append(ref_pos + mic_delta)
    return np.array(unwrapped_pos)

# 封装循环部分：处理有机分子团簇
def process_organic_clusters(structure, new_structure, clusters, is_organic_list):
    """处理有机分子团簇并更新原子位置"""

    for cluster_indices, is_organic in zip(clusters, is_organic_list):
        if is_organic:
            # 解包分子
            unwrapped_pos = unwrap_molecule(structure, cluster_indices)

            # 计算解包后质心
            center_unwrapped = np.mean(unwrapped_pos, axis=0)

            # 将质心转换到分数坐标并映射回晶胞内
            scaled_center = np.dot(center_unwrapped, np.linalg.inv(structure.cell)) % 1.0
            center_original = np.dot(scaled_center, structure.cell)

            # 计算原子相对于质心的位移
            delta_pos = unwrapped_pos - center_unwrapped

            # 在新晶胞中计算新质心
            center_new = np.dot(scaled_center, new_structure.cell)

            # 新位置 = 新质心 + 原始位移
            pos_new = center_new + delta_pos

            # 更新原子位置
            new_structure.positions[cluster_indices] = pos_new
    new_structure.wrap()


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/20 22:22
# @Author  : 兵
# @email    : 1747193328@qq.com
import time

start=time.time()
import numpy as np
from PySide6.QtWidgets import QHBoxLayout, QWidget, QProgressDialog


from NepTrainKit import utils
from NepTrainKit.core import MessageManager, Config
from NepTrainKit.custom_widget import (
    GetIntMessageBox,
    SparseMessageBox,
    IndexSelectMessageBox,
    ShiftEnergyMessageBox,
)
from NepTrainKit.core.io.select import farthest_point_sampling
from NepTrainKit.views.toolbar import NepDisplayGraphicsToolBar
from NepTrainKit.core.energy_shift import shift_dataset_energy, suggest_group_patterns


class NepResultPlotWidget(QWidget):
    def __init__(self,parent=None):
        super().__init__(parent)
        self._parent=parent

        self.draw_mode=False
        # self.setRenderHint(QPainter.Antialiasing, False)
        self._layout = QHBoxLayout(self)
        self.setLayout(self._layout)
        canvas_type = Config.get("widget","canvas_type","pyqtgraph")
        self.last_figure_num=None
        self.swith_canvas(canvas_type)

    def swith_canvas(self,canvas_type="pyqtgraph"):

        if canvas_type == "pyqtgraph":
            from NepTrainKit.core.canvas.pyqtgraph import PyqtgraphCanvas
            self.canvas = PyqtgraphCanvas(self)
            self._layout.addWidget(self.canvas)

        elif canvas_type == "vispy":


            from NepTrainKit.core.canvas.vispy import VispyCanvas


            self.canvas = VispyCanvas(parent=self, bgcolor='white')
            self._layout.addWidget(self.canvas.native)
            # self.window().windowHandle().screenChanged.connect(self.canvas.native.screen_changed)



    # def clear(self):
    #     self.canvas.clear_axes()
        # self.last_figure_num=None

    def set_tool_bar(self, tool):
        self.tool_bar: NepDisplayGraphicsToolBar = tool
        self.tool_bar.panSignal.connect(self.canvas.pan)
        self.tool_bar.resetSignal.connect(self.canvas.auto_range)
        self.tool_bar.deleteSignal.connect(self.canvas.delete)
        self.tool_bar.revokeSignal.connect(self.canvas.revoke)
        self.tool_bar.penSignal.connect(self.canvas.pen)
        self.tool_bar.exportSignal.connect(self.export_descriptor_data)
        self.tool_bar.findMaxSignal.connect(self.find_max_error_point)
        self.tool_bar.discoverySignal.connect(self.find_non_physical_structures)
        self.tool_bar.sparseSignal.connect(self.sparse_point)
        self.tool_bar.shiftEnergySignal.connect(self.shift_energy_baseline)
        self.tool_bar.inverseSignal.connect(self.inverse_select)
        self.tool_bar.selectIndexSignal.connect(self.select_by_index)
        self.canvas.tool_bar=self.tool_bar


    def __find_non_physical_structures(self):
        """
        对每个结构进行非物理距离判断
        """
        structure_list = self.canvas.nep_result_data.structure.now_data
        group_array = self.canvas.nep_result_data.structure.group_array.now_data
        radius_coefficient_config = Config.getfloat("widget","radius_coefficient",0.7)
        unreasonable_index=[]
        for structure,index in zip(structure_list,group_array):

            if not structure.adjust_reasonable(radius_coefficient_config):

                unreasonable_index.append(index)
            yield 1
        self.canvas.select_index(unreasonable_index,False)


    def find_non_physical_structures(self):
        """
        非物理结构函数入口
        """
        if self.canvas.nep_result_data is None:
            return
        progress_diag = QProgressDialog(f"" ,"Cancel",0,self.canvas.nep_result_data.structure.num,self._parent)
        thread=utils.LoadingThread(self._parent,show_tip=False )
        progress_diag.setFixedSize(300, 100)
        progress_diag.setWindowTitle("Finding non-physical structures")
        thread.progressSignal.connect(progress_diag.setValue)
        thread.finished.connect(progress_diag.accept)
        progress_diag.canceled.connect(thread.stop_work)  # 用户取消时终止线程
        thread.start_work(self.__find_non_physical_structures)
        progress_diag.exec()

    def find_max_error_point(self):
        dataset = self.canvas.get_axes_dataset(self.canvas.current_axes)

        if dataset is None:
            return

        box= GetIntMessageBox(self._parent,"Please enter an integer N, it will find the top N structures with the largest errors")
        n = Config.getint("widget","max_error_value",10)
        box.intSpinBox.setValue(n)

        if not box.exec():
            return
        nmax= box.intSpinBox.value()
        Config.set("widget","max_error_value",nmax)
        index= (dataset.get_max_error_index(nmax))

        self.canvas.select_index(index,False)

    def sparse_point(self):
        if  self.canvas.nep_result_data is None:
            return
        box= SparseMessageBox(self._parent,"Please specify the maximum number of structures and minimum distance")
        n_samples = Config.getint("widget","sparse_num_value",10)
        distance = Config.getfloat("widget","sparse_distance_value",0.01)

        box.intSpinBox.setValue(n_samples)
        box.doubleSpinBox.setValue(distance)

        if not box.exec():
            return
        n_samples= box.intSpinBox.value()
        distance= box.doubleSpinBox.value()

        Config.set("widget","sparse_num_value",n_samples)
        Config.set("widget","sparse_distance_value",distance)

        dataset = self.canvas.nep_result_data.descriptor
        if dataset.now_data.size ==0:
            MessageManager.send_message_box("No descriptor data available","Error")
            return
        remaining_indices = farthest_point_sampling(dataset.now_data,n_samples=n_samples,min_dist=distance)

        # 获取所有索引（从 0 到 len(arr)-1）
        # all_indices = np.arange(dataset.now_data.shape[0])

        # 使用 setdiff1d 获取不在 indices_to_remove 中的索引
        # remove_indices = np.setdiff1d(all_indices, remaining_indices)
        structures = dataset.group_array[remaining_indices]
        self.canvas.select_index(structures.tolist(),False)

    def export_descriptor_data(self):
        if self.canvas.nep_result_data is None:
            MessageManager.send_info_message("NEP data has not been loaded yet!")
            return
        path = utils.call_path_dialog(self, "Choose a file save ", "file",default_filename="export_descriptor_data.out")
        if path:
            thread = utils.LoadingThread(self, show_tip=True, title="Exporting descriptor data")
            thread.start_work(self._export_descriptor_data, path)

    def _export_descriptor_data(self,path):

        if len(self.canvas.nep_result_data.select_index) == 0:
            MessageManager.send_info_message("No data selected!")
            return
        select_index=self.canvas.nep_result_data.descriptor.convert_index(list(self.canvas.nep_result_data.select_index))
        descriptor_data = self.canvas.nep_result_data.descriptor.now_data[select_index,:]
        if hasattr(self.canvas.nep_result_data,"energy") and self.canvas.nep_result_data.energy.num !=0:
            select_index = self.canvas.nep_result_data.energy.convert_index(
                list(self.canvas.nep_result_data.select_index))

            energy_data = self.canvas.nep_result_data.energy.now_data[select_index,1]
            descriptor_data = np.column_stack((descriptor_data,energy_data))

        with open(path, "w") as f:
            np.savetxt(f,descriptor_data,fmt='%.6g',delimiter='\t')


    def shift_energy_baseline(self):
        data = self.canvas.nep_result_data
        if data is None:
            return
        ref_index = list(data.select_index)
        # if len(ref_index) == 0:
        #     MessageManager.send_info_message("No data selected!")
        #     return

        max_generations = Config.getint("widget","max_generation_value",100000)
        population_size =  Config.getint("widget","population_size",40)
        convergence_tol = Config.getfloat("widget","convergence_tol", 1e-8)
        config_set = set(data.structure.get_all_config())
        suggested = suggest_group_patterns(list(config_set))
        box = ShiftEnergyMessageBox(
            self._parent,
            "Specify regex groups for Config_type (comma separated)"
        )
        box.groupEdit.setText(";".join(suggested))
        box.genSpinBox.setValue(max_generations)
        box.sizeSpinBox.setValue(population_size)
        box.tolSpinBox.setValue(convergence_tol)


        if not box.exec():
            return

        pattern_text = box.groupEdit.text().strip()
        group_patterns = [p.strip() for p in pattern_text.split(';') if p.strip()]

        alignment_mode = box.modeCombo.currentText()


        max_generations = box.genSpinBox.value()
        population_size = box.sizeSpinBox.value()
        convergence_tol = box.tolSpinBox.value()
        Config.set("widget","max_generation_value",max_generations)
        Config.set("widget","population_size",population_size)
        Config.set("widget","convergence_tol",convergence_tol)
        config_set = set(data.structure.get_all_config())
        progress_diag = QProgressDialog(f"", "Cancel", 0, len(config_set), self._parent)
        thread = utils.LoadingThread(self._parent, show_tip=False)
        progress_diag.setFixedSize(300, 100)
        progress_diag.setWindowTitle("Shift energies")
        thread.progressSignal.connect(progress_diag.setValue)
        thread.finished.connect(progress_diag.accept)
        progress_diag.canceled.connect(thread.stop_work)  # 用户取消时终止线程
        thread.start_work(
            shift_dataset_energy,
            structures=data.structure.now_data,
            reference_structures=data.structure.all_data[ref_index],
            max_generations=max_generations,
            population_size=population_size,
            convergence_tol=convergence_tol,
            group_patterns=group_patterns,
            alignment_mode=alignment_mode,
            nep_energy_array=data.energy.y,
        )
        progress_diag.exec()
        if hasattr(data, "energy") and data.energy.num != 0:
            for i, s in enumerate(data.structure.all_data):
                # print(s.per_atom_energy)
                data.energy.data._data[i, 1] = s.per_atom_energy
        self.canvas.plot_nep_result()

    def inverse_select(self):
        self.canvas.inverse_select()

    def select_by_index(self):
        if self.canvas.nep_result_data is None:
            return
        box = IndexSelectMessageBox(self._parent, "Select structures by index")
        if not box.exec():
            return
        text = box.indexEdit.text().strip()
        use_origin = box.checkBox.isChecked()
        data = self.canvas.nep_result_data.structure
        total = data.all_data.shape[0] if use_origin else data.now_data.shape[0]
        indices = utils.parse_index_string(text, total)
        if not indices:
            return
        if not use_origin:
            indices = data.group_array.now_data[indices].tolist()
        self.canvas.select_index(indices, False)

    def set_dataset(self,dataset):

        if self.last_figure_num !=len(dataset.dataset):

            self.canvas.init_axes(len(dataset.dataset))
            self.last_figure_num = len(dataset.dataset)

        self.canvas.set_nep_result_data(dataset)
        self.canvas.plot_nep_result()


















#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2025/6/18 13:21
# @Author  : 兵
# @email    : 1747193328@qq.com
from itertools import combinations

import numpy as np
from PySide6.QtWidgets import QFrame, QGridLayout
from ase.build import surface
from loguru import logger
from qfluentwidgets import BodyLabel, ComboBox, ToolTipFilter, ToolTipPosition, CheckBox, EditableComboBox

from NepTrainKit.core import CardManager, process_organic_clusters, get_clusters
from NepTrainKit.custom_widget import SpinBoxUnitInputFrame
from NepTrainKit.custom_widget.card_widget import MakeDataCard
from scipy.stats.qmc import Sobol

@CardManager.register_card
class RandomSlabCard(MakeDataCard):
    card_name = "Random Slab"
    menu_icon = r":/images/src/images/supercell.svg"

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle("Random Slab Generation")
        self.init_ui()

    def init_ui(self):
        self.setObjectName("random_slab_card_widget")

        # Miller index ranges for h, k, l
        self.h_label = BodyLabel("h", self.setting_widget)
        self.h_label.setToolTip("h index range")
        self.h_label.installEventFilter(ToolTipFilter(self.h_label, 0, ToolTipPosition.TOP))
        self.h_frame = SpinBoxUnitInputFrame(self)
        self.h_frame.set_input(["-", "step", ""], 3, "int")
        self.h_frame.setRange(-10, 10)
        self.h_frame.set_input_value([0, 1, 1])

        self.k_label = BodyLabel("k", self.setting_widget)
        self.k_label.setToolTip("k index range")
        self.k_label.installEventFilter(ToolTipFilter(self.k_label, 0, ToolTipPosition.TOP))
        self.k_frame = SpinBoxUnitInputFrame(self)
        self.k_frame.set_input(["-", "step", ""], 3, "int")
        self.k_frame.setRange(-10, 10)
        self.k_frame.set_input_value([0, 1, 1])

        self.l_label = BodyLabel("l", self.setting_widget)
        self.l_label.setToolTip("l index range")
        self.l_label.installEventFilter(ToolTipFilter(self.l_label, 0, ToolTipPosition.TOP))
        self.l_frame = SpinBoxUnitInputFrame(self)
        self.l_frame.set_input(["-", "step", ""], 3, "int")
        self.l_frame.setRange(-10, 10)
        self.l_frame.set_input_value([1, 3, 1])

        # Layer number range
        self.layer_label = BodyLabel("Layers", self.setting_widget)
        self.layer_label.setToolTip("Layer range")
        self.layer_label.installEventFilter(ToolTipFilter(self.layer_label, 0, ToolTipPosition.TOP))
        self.layer_frame = SpinBoxUnitInputFrame(self)
        self.layer_frame.set_input(["-", "step", ""], 3, "int")
        self.layer_frame.setRange(1, 50)
        self.layer_frame.set_input_value([3, 6, 1])

        # Vacuum thickness range
        self.vacuum_label = BodyLabel("Vacuum", self.setting_widget)
        self.vacuum_label.setToolTip("Vacuum thickness range in Å")
        self.vacuum_label.installEventFilter(ToolTipFilter(self.vacuum_label, 0, ToolTipPosition.TOP))
        self.vacuum_frame = SpinBoxUnitInputFrame(self)
        self.vacuum_frame.set_input(["-", "step", "Å"], 3, "int")
        self.vacuum_frame.setRange(0, 100)
        self.vacuum_frame.set_input_value([10, 10, 1])

        self.settingLayout.addWidget(self.h_label, 0, 0, 1, 1)
        self.settingLayout.addWidget(self.h_frame, 0, 1, 1, 2)
        self.settingLayout.addWidget(self.k_label, 1, 0, 1, 1)
        self.settingLayout.addWidget(self.k_frame, 1, 1, 1, 2)
        self.settingLayout.addWidget(self.l_label, 2, 0, 1, 1)
        self.settingLayout.addWidget(self.l_frame, 2, 1, 1, 2)
        self.settingLayout.addWidget(self.layer_label, 3, 0, 1, 1)
        self.settingLayout.addWidget(self.layer_frame, 3, 1, 1, 2)
        self.settingLayout.addWidget(self.vacuum_label, 4, 0, 1, 1)
        self.settingLayout.addWidget(self.vacuum_frame, 4, 1, 1, 2)

    def process_structure(self, structure):
        structure_list = []

        h_min, h_max, h_step = self.h_frame.get_input_value()
        k_min, k_max, k_step = self.k_frame.get_input_value()
        l_min, l_max, l_step = self.l_frame.get_input_value()

        layer_min, layer_max, layer_step = self.layer_frame.get_input_value()
        vac_min, vac_max, vac_step = self.vacuum_frame.get_input_value()

        h_range = np.arange(h_min, h_max + 1, h_step)
        k_range = np.arange(k_min, k_max + 1, k_step)
        l_range = np.arange(l_min, l_max + 1, l_step)
        layer_range = np.arange(layer_min, layer_max + 1, layer_step)
        vac_range = np.arange(vac_min, vac_max + vac_step, vac_step)

        for h in h_range:
            for k in k_range:
                for l in l_range:
                    if h == 0 and k == 0 and l == 0:
                        continue
                    for layers in layer_range:
                        for vac in vac_range:
                            try:
                                slab = surface(structure, (int(h), int(k), int(l)), int(layers), vacuum=float(vac))
                                slab.info["Config_type"] = slab.info.get("Config_type", "") + f" Slab(hkl={int(h)}{int(k)}{int(l)},layers={int(layers)},vacuum={vac})"
                                structure_list.append(slab)
                            except Exception as e:
                                logger.error(f"Failed to build slab {(h, k, l)}: {e}")
        return structure_list

    def to_dict(self):
        data_dict = super().to_dict()
        data_dict['h_range'] = self.h_frame.get_input_value()
        data_dict['k_range'] = self.k_frame.get_input_value()
        data_dict['l_range'] = self.l_frame.get_input_value()
        data_dict['layer_range'] = self.layer_frame.get_input_value()
        data_dict['vacuum_range'] = self.vacuum_frame.get_input_value()
        return data_dict

    def from_dict(self, data_dict):
        super().from_dict(data_dict)
        self.h_frame.set_input_value(data_dict.get('h_range', [0, 1, 1]))
        self.k_frame.set_input_value(data_dict.get('k_range', [0, 1, 1]))
        self.l_frame.set_input_value(data_dict.get('l_range', [1, 3, 1]))
        self.layer_frame.set_input_value(data_dict.get('layer_range', [3, 6, 1]))
        self.vacuum_frame.set_input_value(data_dict.get('vacuum_range', [10, 10, 1]))




#这里类名设计失误 但为了兼容以前的配置文件  不再修改类名了



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2025/4/5 20:11
# @Author  : 兵
# @email    : 1747193328@qq.com
from PySide6.QtWidgets import QFrame, QHBoxLayout, QSpinBox, QDoubleSpinBox,QLineEdit
from qfluentwidgets import BodyLabel


class SpinBoxUnitInputFrame(QFrame):
    def __init__(self, parent=None):
        super(SpinBoxUnitInputFrame, self).__init__(parent)
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.object_list = []
    def set_input(self, unit_str,object_num ,input_type="int"):
        if  isinstance(unit_str,str):
            unit_str = [unit_str]*object_num
        elif isinstance(unit_str,list):
            unit_str=unit_str
        else:
            raise TypeError('unit_str must be str or list')

        if  isinstance(input_type,str):
            input_type = [input_type]*object_num
        elif isinstance(input_type,list):
            input_type=input_type
        else:
            raise TypeError('input_type must be str or list')

        for i in range(object_num):
            if input_type[i%len(unit_str)]=="int":
                input_object = QSpinBox(self)
                input_object.setButtonSymbols(QSpinBox.NoButtons)
            elif input_type[i%len(unit_str)]=="float":
                input_object = QDoubleSpinBox(self)
                input_object.setDecimals(3)
                input_object.setButtonSymbols(QSpinBox.NoButtons)
            else:
                raise TypeError('input_type must be int or float')

            input_object.setFixedHeight(25)
            self._layout.addWidget(input_object)
            self._layout.addWidget(BodyLabel(unit_str[i%len(unit_str)],self))
            self.object_list.append(input_object)

    def setRange(self, min_value, max_value):
        for input_object in self.object_list:
            input_object.setRange(min_value, max_value)

    def get_input_value(self):

        return [input_object.value() for input_object in self.object_list]

    def set_input_value(self, value_list):
        if not isinstance(value_list,list):
            value_list=[value_list]*len(self.object_list)

        for i, input_object in enumerate(self.object_list):
            input_object.setValue(value_list[i])


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/12/2 19:58
# @Author  : 兵
# @email    : 1747193328@qq.com
from PySide6.QtCore import Qt, Signal, QSize
from PySide6.QtWidgets import QCompleter
from qfluentwidgets import SearchLineEdit
from qfluentwidgets.components.widgets.line_edit import CompleterMenu, LineEditButton

from .completer import CompleterModel, JoinDelegate


class ConfigTypeSearchLineEdit(SearchLineEdit):
    checkSignal=Signal(str)
    uncheckSignal=Signal(str)
    def __init__(self, parent):
        super().__init__(parent)
        self.init()

    def init(self):
        self.searchButton.setToolTip("Searching for structures based on Config_type")
        self.checkButton = LineEditButton(":/images/src/images/check.svg", self)
        self.checkButton.setToolTip("Mark structure according to Config_type")
        self.uncheckButton = LineEditButton(":/images/src/images/uncheck.svg", self)
        self.uncheckButton.setToolTip("Unmark structure according to Config_type")

        self.searchButton.setIconSize(QSize(16, 16))

        self.checkButton.setIconSize(QSize(16, 16))
        self.uncheckButton.setIconSize(QSize(16, 16))

        self.hBoxLayout.addWidget(self.checkButton, 0, Qt.AlignRight)
        self.hBoxLayout.addWidget(self.uncheckButton, 0, Qt.AlignRight)


        self.checkButton.clicked.connect(self._checked)
        self.uncheckButton.clicked.connect(self._unchecked)


        self.setObjectName("search_lineEdit")
        self.setPlaceholderText("Click to view Config_type")
        stands = []
        self.completer_model = CompleterModel(stands)


        completer = QCompleter( self.completer_model , self)
        completer.setCaseSensitivity(Qt.CaseInsensitive)
        # completer.setMaxVisibleItems(10)
        completer.setFilterMode(Qt.MatchContains)
        self.setCompleter(completer)
        _completerMenu=CompleterMenu(self)
        self.setCompleterMenu(_completerMenu)
        self._delegate =JoinDelegate(self,{})
        _completerMenu.view.setItemDelegate(self._delegate)
        _completerMenu.view.setMaxVisibleItems(10)







    def _checked(self):
        self.checkSignal.emit(self.text())

    def _unchecked(self):
        self.uncheckSignal.emit(self.text())



    def mousePressEvent(self,event):

        self._completer.setCompletionPrefix(self.text())
        self._completerMenu.setCompletion(self._completer.completionModel())
        self._completerMenu.popup()
        super().mousePressEvent(event)


    def setCompleterKeyWord(self, new_words):
        if isinstance(new_words, list):
            new_words = self.completer_model.parser_list(new_words)
        self._delegate.data=new_words
        self.completer_model.set_data(new_words)





#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/18 13:26
# @Author  : 兵
# @email    : 1747193328@qq.com

import os
import traceback
from pathlib import Path

import numpy as np
from PySide6.QtCore import QObject, Signal
from loguru import logger
from NepTrainKit import module_path,utils
from NepTrainKit.core import MessageManager, Structure, Config
from NepTrainKit.core.calculator import NEPProcess



from NepTrainKit.core.io.base import NepPlotData, StructureData


from NepTrainKit.core.io.utils import read_nep_out_file, check_fullbatch, read_nep_in, parse_array_by_atomnum


def pca(X, n_components=None):
    """
    执行主成分分析 (PCA)，只返回降维后的数据
    """
    n_samples, n_features = X.shape

    # 1. 计算均值并中心化数据
    mean = np.mean(X, axis=0)
    X_centered = X - mean
    #樊老师说不用处理 就不减去均值了
    # 但是我还不确定哪种好 还是保持现状把
    # X_centered = X


    # 3. 计算协方差矩阵
    cov_matrix = np.dot(X_centered.T, X_centered) / (n_samples - 1)

    # 4. 计算特征值和特征向量
    eigenvalues, eigenvectors = np.linalg.eigh(cov_matrix)

    # 5. 特征值和特征向量按降序排列
    idx = np.argsort(eigenvalues)[::-1]
    eigenvalues = eigenvalues[idx]
    eigenvectors = eigenvectors[:, idx]

    # 6. 确定要保留的主成分数量
    if n_components is None:
        n_components = n_features
    elif n_components > n_features:
        n_components = n_features

    # 7. 将数据投影到前n_components个主成分上 (降维)
    X_pca = np.dot(X_centered, eigenvectors[:, :n_components])

    return X_pca.astype(np.float32)


class ResultData(QObject):
    #通知界面更新训练集的数量情况
    updateInfoSignal = Signal( )
    loadFinishedSignal = Signal()


    def __init__(self,nep_txt_path,data_xyz_path,descriptor_path):
        super().__init__()
        self.load_flag=False

        self.descriptor_path=descriptor_path
        self.data_xyz_path=data_xyz_path
        self.nep_txt_path=nep_txt_path

        self.select_index=set()

        self.nep_calc_thread = NEPProcess()


    def load_structures(self):
        structures = Structure.read_multiple(self.data_xyz_path)
        self._atoms_dataset=StructureData(structures)
        self.atoms_num_list = np.array([len(struct) for struct in self.structure.now_data])


    def write_prediction(self):
        if self.atoms_num_list.shape[0] > 1000:
            #
            if not self.data_xyz_path.with_name("nep.in").exists():
                with open(self.data_xyz_path.with_name("nep.in"),
                          "w", encoding="utf8") as f:
                    f.write("prediction 1 ")

    def load(self ):
        try:
            self.load_structures()
            self._load_descriptors()
            self._load_dataset()
            self.load_flag=True
        except:
            logger.error(traceback.format_exc())

            MessageManager.send_error_message("load dataset error!")

        self.loadFinishedSignal.emit()
    def _load_dataset(self):
        raise NotImplementedError()

    @property
    def dataset(self) -> ["NepPlotData"]:
        return []

    @property
    def descriptor(self):
        return self._descriptor_dataset

    @property
    def num(self):
        return self._atoms_dataset.num
    @property
    def structure(self):
        return self._atoms_dataset

    def is_select(self,i):
        return i in self.select_index

    def select(self,indices):
        """
        传入一个索引列表，将索引对应的结构标记为选中状态
        这个下标是结构在train.xyz中的索引
        """


        # 统一转换为 NumPy 数组
        idx = np.asarray(indices, dtype=int) if not isinstance(indices, int) else np.array([indices])
        # 去重并过滤有效索引（在数据范围内且为活跃数据）
        idx = np.unique(idx)
        idx = idx[(idx >= 0) & (idx < len(self.structure.all_data)) & (self.structure.data._active_mask[idx])]
        # 批量添加到选中集合
        self.select_index.update(idx)

        self.updateInfoSignal.emit()

    def uncheck(self,_list):
        """
        check_list 传入一个索引列表，将索引对应的结构标记为未选中状态
        这个下标是结构在train.xyz中的索引
        """
        if isinstance(_list,int):
            _list=[_list]
        for i in _list:
            if i in self.select_index:
                self.select_index.remove(i)

        self.updateInfoSignal.emit()

    def inverse_select(self):
        """Invert the current selection state of all active structures"""
        active_indices = set(self.structure.data.now_indices.tolist())
        selected_indices = set(self.select_index)
        unselect = list(selected_indices)
        select = list(active_indices - selected_indices)
        if unselect:
            self.uncheck(unselect)
        if select:
            self.select(select)
    def export_selected_xyz(self,save_file_path):
        """
        导出当前选中的结构
        """
        index=list(self.select_index)
        try:

            with open(save_file_path,"w",encoding="utf8") as f:

                index=self.structure.convert_index(index)

                for structure in self.structure.all_data[index]:
                    structure.write(f)

            MessageManager.send_info_message(f"File exported to: {save_file_path}")
        except:
            MessageManager.send_info_message(f"An unknown error occurred while saving. The error message has been output to the log!")
            logger.error(traceback.format_exc())

    def export_model_xyz(self,save_path):
        """
        导出当前结构
        :param save_path: 保存路径
        被删除的导出到export_remove_model.xyz
        被保留的导出到export_good_model.xyz
        """
        try:

            with open(Path(save_path).joinpath("export_good_model.xyz"),"w",encoding="utf8") as f:
                for structure in self.structure.now_data:
                    structure.write(f)

            with open(Path(save_path).joinpath("export_remove_model.xyz"),"w",encoding="utf8") as f:
                for structure in self.structure.remove_data:
                    structure.write(f)


            MessageManager.send_info_message(f"File exported to: {save_path}")
        except:
            MessageManager.send_info_message(f"An unknown error occurred while saving. The error message has been output to the log!")
            logger.error(traceback.format_exc())


    def get_atoms(self,index ):
        """根据原始索引获取原子结构对象"""
        index=self.structure.convert_index(index)
        return self.structure.all_data[index][0]



    def remove(self,i):

        """
        在所有的dataset中删除某个索引对应的结构
        """
        self.structure.remove(i)
        for dataset in self.dataset:
            dataset.remove(i)
        self.updateInfoSignal.emit()

    @property
    def is_revoke(self):
        """
        判断是否有被删除的结构
        """
        return self.structure.remove_data.size!=0
    def revoke(self):
        """
        撤销到上一次的删除
        """
        self.structure.revoke()
        for dataset in self.dataset:
            dataset.revoke( )
        self.updateInfoSignal.emit()

    @utils.timeit
    def delete_selected(self ):
        """
        删除所有selected的结构
        """
        self.remove(list(self.select_index))
        self.select_index.clear()
        self.updateInfoSignal.emit()


    def _load_descriptors(self):


        if os.path.exists(self.descriptor_path):
            desc_array = read_nep_out_file(self.descriptor_path,dtype=np.float32)

        else:
            desc_array = np.array([])

        if desc_array.size == 0:
            self.nep_calc_thread.run_nep3_calculator_process(self.nep_txt_path.as_posix(),
                self.structure.now_data,
                "descriptor" ,wait=True)
            desc_array=self.nep_calc_thread.func_result
            # desc_array = run_nep3_calculator_process(
            #     )

            if desc_array.size != 0:
                np.savetxt(self.descriptor_path, desc_array, fmt='%.6g')
        else:
            if desc_array.shape[0] == np.sum(self.atoms_num_list):
                # 原子描述符 需要计算结构描述符


                desc_array = parse_array_by_atomnum(desc_array, self.atoms_num_list, map_func=np.mean, axis=0)
            elif desc_array.shape[0] == self.atoms_num_list.shape[0]:
                # 结构描述符
                pass

            else:
                self.descriptor_path.unlink()
                return self._load_descriptors()

        if desc_array.size != 0:
            if desc_array.shape[1] > 2:
                try:
                    desc_array = pca(desc_array, 2)
                except:
                    MessageManager.send_error_message("PCA dimensionality reduction fails")
                    desc_array = np.array([])

        self._descriptor_dataset = NepPlotData(desc_array, title="descriptor")


class NepTrainResultData(ResultData):
    def __init__(self,
                 nep_txt_path,
                 data_xyz_path,
                 energy_out_path,
                 force_out_path,
                 stress_out_path,
                 virial_out_path,
                 descriptor_path

                 ):
        super().__init__(nep_txt_path,data_xyz_path,descriptor_path)
        self.energy_out_path = energy_out_path
        self.force_out_path = force_out_path
        self.stress_out_path = stress_out_path
        self.virial_out_path = virial_out_path

    @property
    def dataset(self):
        # return [self.energy, self.stress,self.virial, self.descriptor]
        return [self.energy,self.force,self.stress,self.virial, self.descriptor]

    @property
    def energy(self):
        return self._energy_dataset

    @property
    def force(self):
        return self._force_dataset

    @property
    def stress(self):
        return self._stress_dataset

    @property
    def virial(self):
        return self._virial_dataset

    @classmethod
    def from_path(cls, path ):
        dataset_path = Path(path)

        file_name=dataset_path.stem

        nep_txt_path = dataset_path.with_name(f"nep.txt")
        if not nep_txt_path.exists():
            nep89_path = os.path.join(module_path, "Config/nep89.txt")
            nep_txt_path=Path(nep89_path)
        energy_out_path = dataset_path.with_name(f"energy_{file_name}.out")
        force_out_path = dataset_path.with_name(f"force_{file_name}.out")
        stress_out_path = dataset_path.with_name(f"stress_{file_name}.out")
        virial_out_path = dataset_path.with_name(f"virial_{file_name}.out")
        if file_name=="train":

            descriptor_path = dataset_path.with_name(f"descriptor.out")
        else:
            descriptor_path = dataset_path.with_name(f"descriptor_{file_name}.out")
        return cls(nep_txt_path,dataset_path,energy_out_path,force_out_path,stress_out_path,virial_out_path,descriptor_path)

    def _load_dataset(self) -> None:
        """加载或计算 NEP 数据集，并更新内部数据集属性。"""
        nep_in = read_nep_in(self.data_xyz_path.with_name("nep.in"))
        if self._should_recalculate(nep_in):
            energy_array, force_array, virial_array, stress_array = self._recalculate_and_save( )
        else:
            energy_array = read_nep_out_file(self.energy_out_path, dtype=np.float32)
            force_array = read_nep_out_file(self.force_out_path, dtype=np.float32)
            virial_array = read_nep_out_file(self.virial_out_path, dtype=np.float32)
            stress_array = read_nep_out_file(self.stress_out_path, dtype=np.float32)

            if energy_array.shape[0]!=self.atoms_num_list.shape[0]:
                self.energy_out_path.unlink(True)
                self.force_out_path.unlink(True)
                self.virial_out_path.unlink(True)
                self.stress_out_path.unlink(True)

                return self._load_dataset()


        self._energy_dataset = NepPlotData(energy_array, title="energy")
        default_forces = Config.get("widget", "forces_data", "Row")
        if force_array.size != 0 and default_forces == "Norm":

            force_array = parse_array_by_atomnum(force_array, self.atoms_num_list, map_func=np.linalg.norm, axis=0)

            self._force_dataset = NepPlotData(force_array, title="force")
        else:
            self._force_dataset = NepPlotData(force_array, group_list=self.atoms_num_list, title="force")

        if float(nep_in.get("lambda_v", 1)) != 0:
            self._stress_dataset = NepPlotData(stress_array, title="stress")

            self._virial_dataset = NepPlotData(virial_array, title="virial")
        else:
            self._stress_dataset = NepPlotData([], title="stress")

            self._virial_dataset = NepPlotData([], title="virial")
    def _should_recalculate(self, nep_in: dict) -> bool:
        """判断是否需要重新计算 NEP 数据。"""
        output_files_exist = all([
            self.energy_out_path.exists(),
            self.force_out_path.exists(),
            self.stress_out_path.exists(),
            self.virial_out_path.exists()
        ])
        return not check_fullbatch(nep_in, len(self.atoms_num_list)) or not output_files_exist

    def _save_energy_data(self, potentials: np.ndarray)  :

        """保存能量数据到文件。"""

        try:
            ref_energies = np.array([s.per_atom_energy for s in self.structure.now_data], dtype=np.float32)

            if potentials.size  == 0:
                #计算失败 空数组
                energy_array = np.column_stack([ref_energies, ref_energies])
            else:
                energy_array = np.column_stack([potentials / self.atoms_num_list, ref_energies])
        except Exception:
            logger.debug(traceback.format_exc())
            if potentials.size == 0:
                # 计算失败 空数组
                energy_array = np.column_stack([potentials, potentials])
            else:
                energy_array = np.column_stack([potentials / self.atoms_num_list, potentials / self.atoms_num_list])
        energy_array = energy_array.astype(np.float32)
        if energy_array.size != 0:
            np.savetxt(self.energy_out_path, energy_array, fmt='%10.8f')
        return energy_array

    def _save_force_data(self, forces: np.ndarray)  :
        """保存力数据到文件。"""
        try:
            ref_forces = np.vstack([s.forces for s in self.structure.now_data], dtype=np.float32)

            if forces.size == 0:
                # 计算失败 空数组
                forces_array = np.column_stack([ref_forces, ref_forces])

            else:
                forces_array = np.column_stack([forces, ref_forces])
        except KeyError:
            MessageManager.send_warning_message("use nep3 calculator to calculate forces replace the original forces")
            forces_array = np.column_stack([forces, forces])

        except Exception:
            logger.debug(traceback.format_exc())
            forces_array = np.column_stack([forces, forces])
            MessageManager.send_error_message("an error occurred while calculating forces. Please check the input file.")
        if forces_array.size != 0:
            np.savetxt(self.force_out_path, forces_array, fmt='%10.8f')


        return forces_array



    def _save_virial_and_stress_data(self, virials: np.ndarray )    :
        """保存维里张量和应力数据到文件。"""
        coefficient = (self.atoms_num_list / np.array([s.volume for s in self.structure.now_data]))[:, np.newaxis]
        try:
            ref_virials = np.vstack([s.nep_virial for s in self.structure.now_data], dtype=np.float32)
            if virials.size == 0:
                # 计算失败 空数组
                virials_array = np.column_stack([ref_virials, ref_virials])
            else:
                virials_array = np.column_stack([virials, ref_virials])
        except AttributeError:
            MessageManager.send_warning_message("use nep3 calculator to calculate virial replace the original virial")
            virials_array = np.column_stack([virials, virials])

        except Exception:
            MessageManager.send_error_message(f"An error occurred while calculating virial and stress. Please check the input file.")
            logger.debug(traceback.format_exc())
            virials_array = np.column_stack([virials, virials])

        stress_array = virials_array * coefficient  * 160.21766208  # 单位转换\

        stress_array = stress_array.astype(np.float32)
        if virials_array.size != 0:
            np.savetxt(self.virial_out_path, virials_array, fmt='%10.8f')
        if stress_array.size != 0:
            np.savetxt(self.stress_out_path, stress_array, fmt='%10.8f')


        return virials_array, stress_array

    def _recalculate_and_save(self ):

        try:
            self.nep_calc_thread.run_nep3_calculator_process(self.nep_txt_path.as_posix(),
                self.structure.now_data,
                "calculate" ,wait=True)
            nep_potentials_array, nep_forces_array, nep_virials_array=self.nep_calc_thread.func_result
            # nep_potentials_array, nep_forces_array, nep_virials_array = run_nep3_calculator_process(
            #     self.nep_txt_path.as_posix(),
            #     self.structure.now_data,"calculate")
            if nep_potentials_array.size == 0:
                MessageManager.send_warning_message("The nep calculator fails to calculate the potentials, use the original potentials instead.")


            energy_array = self._save_energy_data(nep_potentials_array)
            force_array = self._save_force_data(nep_forces_array)
            virial_array, stress_array = self._save_virial_and_stress_data(nep_virials_array)


            self.write_prediction()
            return energy_array,force_array,virial_array, stress_array
        except Exception as e:
            logger.debug(traceback.format_exc())
            MessageManager.send_error_message(f"An error occurred while running NEP3 calculator: {e}")
            return np.array([]), np.array([]), np.array([]), np.array([])









class NepPolarizabilityResultData(ResultData):
    def __init__(self,
                 nep_txt_path,
                 data_xyz_path,
                 polarizability_out_path,

        descriptor_path
                 ):
        super().__init__(nep_txt_path,data_xyz_path,descriptor_path)
        self.polarizability_out_path = polarizability_out_path

    @property
    def dataset(self):

        return [self.polarizability_diagonal,self.polarizability_no_diagonal, self.descriptor]



    @property
    def polarizability_diagonal(self):
        return self._polarizability_diagonal_dataset
    @property
    def polarizability_no_diagonal(self):
        return self._polarizability_no_diagonal_dataset

    @property
    def descriptor(self):
        return self._descriptor_dataset

    @classmethod
    def from_path(cls, path ):
        dataset_path = Path(path)
        file_name = dataset_path.stem
        nep_txt_path = dataset_path.with_name(f"nep.txt")
        polarizability_out_path = dataset_path.with_name(f"polarizability_{file_name}.out")
        if file_name == "train":
            descriptor_path = dataset_path.with_name(f"descriptor.out")
        else:
            descriptor_path = dataset_path.with_name(f"descriptor_{file_name}.out")

        return cls(nep_txt_path, dataset_path, polarizability_out_path, descriptor_path)
    def _should_recalculate(self, nep_in: dict) -> bool:
        """判断是否需要重新计算 NEP 数据。"""
        output_files_exist = all([
            self.polarizability_out_path.exists(),

        ])
        return not check_fullbatch(nep_in, len(self.atoms_num_list)) or not output_files_exist

    def _recalculate_and_save(self ):

        try:
            # nep_polarizability_array = run_nep3_calculator_process(self.nep_txt_path.as_posix(),
            #                                                        self.structure.now_data, "polarizability")
            self.nep_calc_thread.run_nep3_calculator_process(self.nep_txt_path.as_posix(),
                self.structure.now_data,
                "polarizability" ,wait=True)
            nep_polarizability_array=self.nep_calc_thread.func_result
            if nep_polarizability_array.size == 0:
                MessageManager.send_warning_message("The nep calculator fails to calculate the polarizability, use the original polarizability instead.")
            nep_polarizability_array = self._save_polarizability_data(  nep_polarizability_array)
            self.write_prediction()

        except Exception as e:
            logger.debug(traceback.format_exc())
            MessageManager.send_error_message(f"An error occurred while running NEP3 calculator: {e}")

            nep_polarizability_array = np.array([])
        return nep_polarizability_array
    def _save_polarizability_data(self, polarizability: np.ndarray)  :
        """保存polarizability数据到文件。"""
        nep_polarizability_array = polarizability / (self.atoms_num_list[:, np.newaxis])

        try:
            ref_polarizability = np.vstack([s.nep_polarizability for s in self.structure.now_data], dtype=np.float32)
            if polarizability.size == 0:
                # 计算失败 空数组
                polarizability_array = np.column_stack([ref_polarizability, ref_polarizability])
            else:

                polarizability_array = np.column_stack([nep_polarizability_array,
                                                        ref_polarizability

                                                        ])

        except Exception:
            logger.debug(traceback.format_exc())
            polarizability_array = np.column_stack([polarizability, polarizability])
        polarizability_array = polarizability_array.astype(np.float32)
        if polarizability_array.size != 0:
            np.savetxt(self.polarizability_out_path, polarizability_array, fmt='%10.8f')

        return polarizability_array

    def _load_dataset(self) -> None:
        """加载或计算 NEP 数据集，并更新内部数据集属性。"""
        nep_in = read_nep_in(self.data_xyz_path.with_name("nep.in"))
        if self._should_recalculate(nep_in):
            polarizability_array = self._recalculate_and_save( )
        else:
            polarizability_array= read_nep_out_file(self.polarizability_out_path, dtype=np.float32)
            if polarizability_array.shape[0]!=self.atoms_num_list.shape[0]:
                self.polarizability_out_path.unlink()
                return self._load_dataset()
        self._polarizability_diagonal_dataset = NepPlotData(polarizability_array[:, [0,1,2,6,7,8]], title="Polar Diag")

        self._polarizability_no_diagonal_dataset = NepPlotData(polarizability_array[:, [3,4,5,9,10,11]], title="Polar NoDiag")


class NepDipoleResultData(ResultData):
    def __init__(self,
                 nep_txt_path,
                 data_xyz_path,
                 dipole_out_path,

                 descriptor_path
                 ):
        super().__init__(nep_txt_path, data_xyz_path, descriptor_path)

        self.dipole_out_path = dipole_out_path
    @property
    def dataset(self):
        return [self.dipole , self.descriptor]

    @property
    def dipole(self):
        return self._dipole_dataset



    @property
    def descriptor(self):
        return self._descriptor_dataset

    @classmethod
    def from_path(cls, path, model="train"):
        dataset_path = Path(path)

        file_name = dataset_path.stem

        nep_txt_path = dataset_path.with_name(f"nep.txt")

        polarizability_out_path = dataset_path.with_name(f"dipole_{file_name}.out")

        if file_name == "train":

            descriptor_path = dataset_path.with_name(f"descriptor.out")
        else:
            descriptor_path = dataset_path.with_name(f"descriptor_{file_name}.out")

        return cls(nep_txt_path, dataset_path, polarizability_out_path, descriptor_path)


    def _should_recalculate(self, nep_in: dict) -> bool:
        """判断是否需要重新计算 NEP 数据。"""


        output_files_exist = all([
            self.dipole_out_path.exists(),

        ])
        return not check_fullbatch(nep_in, len(self.atoms_num_list)) or not output_files_exist

    def _recalculate_and_save(self ):

        try:
            # nep_dipole_array = run_nep3_calculator_process(self.nep_txt_path.as_posix(),
            #                                                self.structure.now_data, "dipole")
            self.nep_calc_thread.run_nep3_calculator_process(self.nep_txt_path.as_posix(),
                self.structure.now_data,
                "dipole" ,wait=True)
            nep_dipole_array=self.nep_calc_thread.func_result

            if nep_dipole_array.size == 0:
                MessageManager.send_warning_message("The nep calculator fails to calculate the dipole, use the original dipole instead.")
            nep_dipole_array = self._save_dipole_data(  nep_dipole_array)
            self.write_prediction()

        except Exception as e:
            logger.debug(traceback.format_exc())
            MessageManager.send_error_message(f"An error occurred while running NEP3 calculator: {e}")

            nep_dipole_array = np.array([])
        return nep_dipole_array
    def _save_dipole_data(self, dipole: np.ndarray)  :
        """保存dipole数据到文件。"""
        nep_dipole_array = dipole / (self.atoms_num_list[:, np.newaxis])

        try:
            ref_dipole = np.vstack([s.nep_dipole for s in self.structure.now_data], dtype=np.float32)
            if dipole.size == 0:
                # 计算失败 空数组
                dipole_array = np.column_stack([ref_dipole, ref_dipole])
            else:
                dipole_array = np.column_stack([nep_dipole_array,
                                            ref_dipole

                                                    ])

        except Exception:
            logger.debug(traceback.format_exc())
            dipole_array = np.column_stack([nep_dipole_array, nep_dipole_array])
        dipole_array = dipole_array.astype(np.float32)
        if dipole_array.size != 0:
            np.savetxt(self.dipole_out_path, dipole_array, fmt='%10.8f')

        return dipole_array

    def _load_dataset(self) -> None:
        """加载或计算 NEP 数据集，并更新内部数据集属性。"""
        nep_in = read_nep_in(self.data_xyz_path.with_name("nep.in"))
        if self._should_recalculate(nep_in):
            dipole_array = self._recalculate_and_save( )
        else:
            dipole_array= read_nep_out_file(self.dipole_out_path, dtype=np.float32)
            if dipole_array.shape[0]!=self.atoms_num_list.shape[0]:
                self.dipole_out_path.unlink()
                return self._load_dataset()
        self._dipole_dataset = NepPlotData(dipole_array, title="dipole")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2025/6/18 13:21
# @Author  : 兵
# @email    : 1747193328@qq.com
from itertools import combinations

import numpy as np
from PySide6.QtWidgets import QFrame, QGridLayout
from qfluentwidgets import BodyLabel, ComboBox, ToolTipFilter, ToolTipPosition, CheckBox, EditableComboBox

from NepTrainKit.core import CardManager, process_organic_clusters, get_clusters
from NepTrainKit.custom_widget import SpinBoxUnitInputFrame
from NepTrainKit.custom_widget.card_widget import MakeDataCard
@CardManager.register_card

class CellStrainCard(MakeDataCard):
    card_name= "Lattice Strain"
    menu_icon=r":/images/src/images/scaling.svg"
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle("Make Cell Strain")

        self.init_ui()

    def init_ui(self):
        self.setObjectName("cell_strain_card_widget")


        self.engine_label=BodyLabel("Axes:",self.setting_widget)
        self.engine_type_combo=EditableComboBox(self.setting_widget)
        axes_type=["uniaxial","biaxial","triaxial","isotropic"]
        self.engine_type_combo.addItems(axes_type)
        self.engine_label.setToolTip('Pull down to select or enter a specific axis, such as X or XY')
        self.engine_label.installEventFilter(ToolTipFilter(self.engine_label, 300, ToolTipPosition.TOP))

        self.optional_frame=QFrame(self.setting_widget)
        self.optional_frame_layout = QGridLayout(self.optional_frame)
        self.optional_frame_layout.setContentsMargins(0,0,0,0)
        self.optional_frame_layout.setSpacing(2)

        self.optional_label=BodyLabel("Optional",self.setting_widget)
        self.organic_checkbox=CheckBox("Identify organic", self.setting_widget)
        self.organic_checkbox.setChecked(True)
        self.organic_checkbox.setToolTip("Treat organic molecules as rigid units")
        self.organic_checkbox.installEventFilter(ToolTipFilter(self.organic_checkbox, 300, ToolTipPosition.TOP))


        self.optional_frame_layout.addWidget(self.organic_checkbox,0,0,1,1)

        self.strain_x_label=BodyLabel("X:",self.setting_widget)
        self.strain_x_frame = SpinBoxUnitInputFrame(self)
        self.strain_x_frame.set_input(["-","% step:","%"],3,"float")
        self.strain_x_frame.setRange(-100,100)

        self.strain_x_frame.set_input_value([-5,5,1])
        self.strain_x_label.setToolTip("X-axis strain range")
        self.strain_x_label.installEventFilter(ToolTipFilter(self.strain_x_label, 300, ToolTipPosition.TOP))

        self.strain_y_label=BodyLabel("Y:",self.setting_widget)
        self.strain_y_frame = SpinBoxUnitInputFrame(self)
        self.strain_y_frame.set_input(["-","% step:","%"],3,"float")
        self.strain_y_frame.setRange(-100,100)
        self.strain_y_frame.set_input_value([-5,5,1])
        self.strain_y_label.setToolTip("Y-axis strain range")
        self.strain_y_label.installEventFilter(ToolTipFilter(self.strain_y_label, 300, ToolTipPosition.TOP))

        self.strain_z_label=BodyLabel("Z:",self.setting_widget)
        self.strain_z_frame = SpinBoxUnitInputFrame(self)
        self.strain_z_frame.set_input(["-","% step:","%"],3,"float")
        self.strain_z_frame.setRange(-100,100)
        self.strain_z_frame.set_input_value([-5,5,1])
        self.strain_z_label.setToolTip("Z-axis strain range")
        self.strain_z_label.installEventFilter(ToolTipFilter(self.strain_z_label, 300, ToolTipPosition.TOP))

        self.settingLayout.addWidget(self.engine_label,0, 0,1, 1)
        self.settingLayout.addWidget(self.engine_type_combo,0, 1, 1, 2)
        self.settingLayout.addWidget(self.optional_label, 1, 0, 1, 1)
        self.settingLayout.addWidget(self.optional_frame, 1, 1, 1,1)
        self.settingLayout.addWidget(self.strain_x_label, 2, 0, 1, 1)
        self.settingLayout.addWidget(self.strain_x_frame, 2, 1, 1,1)
        self.settingLayout.addWidget(self.strain_y_label, 3, 0, 1, 1)
        self.settingLayout.addWidget(self.strain_y_frame, 3, 1, 1,1)
        self.settingLayout.addWidget(self.strain_z_label, 4, 0, 1, 1)
        self.settingLayout.addWidget(self.strain_z_frame, 4, 1, 1,1)

    def process_structure(self, structure):
        structure_list=[]
        axes=self.engine_type_combo.currentText()
        x=self.strain_x_frame.get_input_value()
        y=self.strain_y_frame.get_input_value()
        z=self.strain_z_frame.get_input_value()
        identify_organic=self.organic_checkbox.isChecked()


        if identify_organic:
            clusters, is_organic_list = get_clusters(structure)
        strain_range=[
            np.arange(start=x[0],stop=x[1]+1,step=x[2]),
            np.arange(start=y[0], stop=y[1]+1, step=y[2]),
            np.arange(start=z[0], stop=z[1]+1, step=z[2]),
        ]
        cell = structure.get_cell()
        # Define possible axes (0: x, 1: y, 2: z)
        all_axes = [0, 1, 2]


        if axes == 'isotropic':
            for strain in strain_range[0]:
                new_structure = structure.copy()
                new_cell = cell.copy() * (1 + strain / 100)
                new_structure.set_cell(new_cell, scale_atoms=True)
                if identify_organic:
                    process_organic_clusters(structure, new_structure, clusters, is_organic_list )

                strain_info = [f"all:{strain}%"]
                new_structure.info["Config_type"] = new_structure.info.get("Config_type", "") + f" Strain({'|'.join(strain_info)})"
                structure_list.append(new_structure)
        else:
            if axes == 'uniaxial':
                axes_combinations = [[i] for i in all_axes]
            elif axes == 'biaxial':
                axes_combinations = list(combinations(all_axes, 2))
            elif axes == 'triaxial':
                axes_combinations = [all_axes]
            else:
                axes_combinations = [["XYZ".index(i.upper()) for i in axes if i.upper() in "XYZ"]]
            for ax_comb in axes_combinations:
                if len(ax_comb) == 0:
                    continue
                strain_combinations = (np.array(np.meshgrid(*[strain_range[_] for _ in ax_comb])).T.reshape(-1, len(ax_comb)))
                for strain_vals in strain_combinations:
                    new_structure = structure.copy()
                    new_cell = cell.copy()
                    for ax_idx, strain in zip(ax_comb, strain_vals):
                        new_cell[ax_idx] *= (1 + strain / 100)
                    new_structure.set_cell(new_cell, scale_atoms=True)
                    if identify_organic:
                        process_organic_clusters(structure, new_structure, clusters, is_organic_list)

                    strain_info = ["XYZ"[ax] + ":" + str(s) + "%" for ax, s in zip(ax_comb, strain_vals)]
                    new_structure.info["Config_type"] = new_structure.info.get("Config_type", "") + f" Strain({'|'.join(strain_info)})"
                    structure_list.append(new_structure)

        return structure_list

    def to_dict(self):
        data_dict = super().to_dict()
        data_dict['organic'] = self.organic_checkbox.isChecked()

        data_dict['engine_type'] = self.engine_type_combo.currentText()
        data_dict['x_range'] = self.strain_x_frame.get_input_value()
        data_dict['y_range'] = self.strain_y_frame.get_input_value()
        data_dict['z_range'] = self.strain_z_frame.get_input_value()

        return data_dict

    def from_dict(self, data_dict):
        super().from_dict(data_dict)

        self.organic_checkbox.setChecked(data_dict.get("organic", False))

        self.engine_type_combo.setText(data_dict['engine_type'])

        self.strain_x_frame.set_input_value(data_dict['x_range'])
        self.strain_y_frame.set_input_value(data_dict['y_range'])
        self.strain_z_frame.set_input_value(data_dict['z_range'])






#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/28 12:52
# @Author  :
# @email    : 1747193328@qq.com

import os
import platform
import sys
from loguru import logger
sys.path.append(os.path.dirname(os.path.abspath(__file__)))
import src_rc

try:
    # Actual if statement not needed, but keeps code inspectors more happy
    if __nuitka_binary_dir is not None:
        is_nuitka_compiled = True
except NameError:
    is_nuitka_compiled = False



if is_nuitka_compiled:


    logger.add("./Log/{time:%Y-%m}.log",
               level="DEBUG",
                )
    module_path="./"
else:

    module_path = os.path.dirname(__file__)

def get_user_config_path():
    if platform.system() == 'Windows':
        # Windows 系统通常使用 AppData 路径存放应用数据
        local_path = os.getenv('LOCALAPPDATA', None)
        if local_path is None:
            local_path = os.getenv('USERPROFILE', '') + '\\AppData\\Local '
        user_config_path = os.path.join(local_path, 'NepTrainKit')
    else:
        user_config_path = os.path.expanduser("~/.config/NepTrainKit")
    return user_config_path



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2025/4/7 14:06
# @Author  : 兵
# @email    : 1747193328@qq.com
from qfluentwidgets import  BodyLabel

from PySide6.QtGui import QPainter, QLinearGradient, QColor
from PySide6.QtCore import Qt, QRectF

class ProcessLabel(BodyLabel):
    def __init__(self, parent=None):
        super(ProcessLabel, self).__init__(parent)

        self._progress = 0  # 0~100
        self.set_colors( ["white" ])
    def set_progress(self, value):
        """设置进度(0-100)"""
        self._progress = max(0, min(100, value))
        self.update()  # 触发重绘
    def set_colors(self, colors):
        """设置渐变颜色列表(QColor对象或十六进制字符串)"""
        if len(colors)==1:
            colors=[colors[0] ]*2

        self._colors = [QColor(c) if isinstance(c, str) else c for c in colors]

        self.update()

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        # 计算进度对应的宽度
        progress_width = self.width() * (self._progress / 100)

        # 创建水平渐变
        gradient = QLinearGradient(0, 0, self.width(), 0)
        for i, color in enumerate(self._colors):
            gradient.setColorAt(i / (len(self._colors) - 1), color)

        # 绘制背景（全宽度，半透明）
        painter.setBrush(gradient)
        painter.setPen(Qt.NoPen)
        painter.drawRect(QRectF(0, 0, self.width(), self.height()))

        # 绘制进度遮罩（右侧未完成部分用灰色覆盖）
        if self._progress < 100:
            mask_color = QColor("white")
            painter.setBrush(mask_color)
            painter.drawRect(QRectF(progress_width, 0,
                                    self.width() - progress_width, self.height()))

        # 可选：绘制进度文本
        painter.setPen(Qt.black)
        painter.drawText(self.rect(), Qt.AlignCenter,self.text())


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/21 14:22
# @Author  : 兵
# @email    : 1747193328@qq.com
import contextlib
import os
import traceback
import numpy as np
from PySide6.QtCore import QThread, Signal, QObject
from loguru import logger
from multiprocessing import Process, JoinableQueue, Event

from NepTrainKit import utils
from NepTrainKit.core import Structure, MessageManager

try:
    from NepTrainKit.nep_cpu import CpuNep
except ImportError:
    logger.debug(traceback.format_exc())
    try:
        from nep_cpu import CpuNep
    except ImportError:
        logger.debug(traceback.format_exc())

        CpuNep=None

class NepCalculator():

    def __init__(self, model_file="nep.txt"):
        super().__init__()
        if not isinstance(model_file, str):
            model_file = str(model_file )
        self.initialized = False

        if CpuNep is None:
            MessageManager.send_message_box("Failed to import nep_cpu.\n To use the display functionality normally, please prepare the *_train.out and descriptor.out files.","Error")
            return

        if os.path.exists(model_file):

            with open(os.devnull, 'w') as devnull:
                with contextlib.redirect_stdout(devnull), contextlib.redirect_stderr(devnull):
                    self.nep3 = CpuNep(model_file)
            self.element_list=self.nep3.get_element_list()
            self.type_dict = {e: i for i, e in enumerate(self.element_list)}

            self.initialized = True

    def compose_structures(self, structures:list[Structure]):
        group_size = []
        _types = []
        _boxs = []
        _positions = []
        if  not isinstance(structures,(list,np.ndarray)):
            structures = [structures]
        for structure in structures:
            symbols = structure.get_chemical_symbols()
            _type = [self.type_dict[k] for k in symbols]
            _box = structure.cell.transpose(1, 0).reshape(-1).tolist()
            _position = structure.positions.transpose(1, 0).reshape(-1).tolist()
            _types.append(_type)
            _boxs.append(_box)
            _positions.append(_position)
            group_size.append(len(_type))
        return  _types, _boxs, _positions,group_size

    @utils.timeit
    def calculate(self,structures:list[Structure]):
        if not self.initialized:
            return np.array([]),np.array([]),np.array([])
        _types, _boxs, _positions,group_size = self.compose_structures(structures)
        potentials, forces, virials = self.nep3.calculate(_types, _boxs, _positions)
        split_indices = np.cumsum(group_size)[:-1]
        #
        potentials=np.hstack(potentials)
        split_potential_arrays = np.split(potentials, split_indices)
        potentials_array = np.array(list(map(np.sum, split_potential_arrays)), dtype=np.float32)
        # print(potentials_array)
        # 处理每个force数组：reshape (3, -1) 和 transpose(1, 0)
        reshaped_forces = [np.array(force).reshape(3, -1).T for force in forces]

        forces_array = np.vstack(reshaped_forces,dtype=np.float32)
        # print(forces_array)

        reshaped_virials = np.vstack([np.array(virial).reshape(9, -1).mean(axis=1) for virial in virials],dtype=np.float32)

        virials_array = reshaped_virials[:,[0,4,8,1,5,6]]

        return potentials_array,forces_array,virials_array


    def get_descriptor(self,structure:Structure):
        """
        获取单个结构的所有原子描述符
        """
        if not self.initialized:
            return np.array([])
        symbols = structure.get_chemical_symbols()
        _type = [self.type_dict[k] for k in symbols]
        _box = structure.cell.transpose(1, 0).reshape(-1).tolist()

        _position = structure.positions.transpose(1, 0).reshape(-1).tolist()

        descriptor = self.nep3.get_descriptor(_type, _box, _position)

        descriptors_per_atom = np.array(descriptor,dtype=np.float32).reshape(-1, len(structure)).T

        return descriptors_per_atom
    @utils.timeit
    def get_structures_descriptor(self,structures:list[Structure]):
        """
        获取结构描述符：原子平均
        返回的已经结构的描述符了 无需平均
        """
        if not self.initialized:
            return np.array([])
        _types, _boxs, _positions, group_size = self.compose_structures(structures)

        descriptor = self.nep3.get_structures_descriptor(_types, _boxs, _positions)

        return np.array(descriptor,dtype=np.float32)

    @utils.timeit
    def get_structures_polarizability(self,structures:list[Structure]):
        if not self.initialized:
            return np.array([])
        _types, _boxs, _positions, group_size = self.compose_structures(structures)
        polarizability = self.nep3.get_structures_polarizability(_types, _boxs, _positions)

        return np.array(polarizability,dtype=np.float32)

    def get_structures_dipole(self,structures:list[Structure]):
        if not self.initialized:
            return np.array([])
        _types, _boxs, _positions, group_size = self.compose_structures(structures)

        dipole = self.nep3.get_structures_dipole(_types, _boxs, _positions)

        return np.array(dipole,dtype=np.float32)

Nep3Calculator = NepCalculator

def run_nep_calculator(nep_txt, structures, calculator_type, queue=None):
    try:

        nep3 = NepCalculator(nep_txt)
        if calculator_type == 'polarizability':
            result = nep3.get_structures_polarizability(structures)
        elif calculator_type == 'descriptor':
            result = nep3.get_structures_descriptor(structures)
        elif calculator_type == 'dipole':
            result = nep3.get_structures_dipole(structures)
        else:
            result = nep3.calculate(structures)
        if queue:
            queue.put(result)  # 将结果通过管道发送给主进程
    except Exception as e:
        logger.error(traceback.format_exc())
        result = np.array([])
        if queue:
            queue.put(result)
    if   queue is  None:
        return result

class NEPProcess(QObject):
    result_signal = Signal( )  # 信号，用于传递进程结果
    error_signal = Signal(str )  # 信号，用于传递错误信息

    def __init__(self ):
        super().__init__()

        self.process = None
        self.use_process = False
        self.func_result = None
    def run_nep3_calculator_process(self,nep_txt, structures, calculator_type="calculate",wait=False):
        self.func=run_nep_calculator
        self.queue = JoinableQueue()

        self.input_kwargs={
            "nep_txt": nep_txt,
            "calculator_type": calculator_type,
            "structures": structures,

        }
        if len(structures) < 2000:
            self.use_process=False
            self.input_kwargs["queue"] =  None
        else:
            self.input_kwargs["queue"] = self.queue
            self.use_process=True

        self.input_args=()
        self.func_result=None
        self.run()

    def run(self):
        try:
            if self.use_process:
                # 创建并启动进程
                self.process = Process(target=self.func, args= self.input_args ,kwargs = self.input_kwargs,daemon=True)
                self.process.start()
                # 获取结果
                self.func_result = self.queue.get()
            else:
                self.func_result=self.func(*self.input_args,**self.input_kwargs)
            if self.func_result is not None:
                self.result_signal.emit( )
            else:
                self.error_signal.emit("No result returned from process")

        except Exception as e:
            logger.error(traceback.format_exc())
            self.error_signal.emit(f"Error: {str(e)}")
        finally:
            self.queue.close()  # 关闭队列
            # self.queue.join()  # 等待队列任务完成
            if self.process is not None :
                self.process.terminate()  # 确保进程被终止

                self.process.join()
                # self.process.close()

    def stop(self):
        """强制终止进程并清理资源"""
        try:
            if self.process is not None:
                if self.process.is_alive():

                    self.process.terminate()

                    self.process.join()
            self.queue.cancel_join_thread()
        except Exception as e:
            logger.error(traceback.format_exc())

if __name__ == '__main__':
    structures = Structure.read_multiple(r"D:\Desktop\nep\nep-data-main\2023_Zhao_PdCuNiP\train.xyz")
    nep = NepCalculator(r"D:\Desktop\nep\nep-data-main\2023_Zhao_PdCuNiP\nep.txt")
    nep.calculate(structures)


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/17 17:21
# @Author  : 兵
# @email    : 1747193328@qq.com
from .makedata import MakeDataWidget
from .settings import SettingsWidget
from .show_nep import ShowNepWidget
